##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2022-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import codecs
from datetime import datetime, timedelta
from fractions import Fraction
import logging
import math
import pprint
import re

from photini.exiv2 import MetadataHandler
from photini.pyqt import *
from photini.pyqt import qt_version_info, using_pyside

logger = logging.getLogger(__name__)

# photini.metadata imports these classes
__all__ = (
    'MD_Aperture', 'MD_CameraModel', 'MD_ContactInformation', 'MD_DateTime',
    'MD_Dimensions', 'MD_GPSinfo', 'MD_ImageRegion', 'MD_Int', 'MD_Keywords',
    'MD_LangAlt', 'MD_LensModel', 'MD_MultiLocation', 'MD_MultiString',
    'MD_Orientation', 'MD_Rating', 'MD_Rational', 'MD_Rights',
    'MD_SingleLocation', 'MD_Software', 'MD_String', 'MD_Thumbnail',
    'MD_Timezone', 'MD_VideoDuration', 'safe_fraction')


def safe_fraction(value, limit=True):
    # Avoid ZeroDivisionError when '0/0' used for zero values in Exif
    try:
        if isinstance(value, (list, tuple)):
            value = Fraction(*value)
        else:
            value = Fraction(value)
    except ZeroDivisionError:
        return Fraction(0.0)
    if limit:
        # round off excessively large denominators
        value = value.limit_denominator(1000000)
    return value


class MD_Value(object):
    # mixin for "metadata objects" - Python types with additional functionality
    _quiet = False

    @classmethod
    def from_ffmpeg(cls, file_value, tag):
        return cls(file_value)

    @classmethod
    def from_exiv2(cls, file_value, tag):
        return cls(file_value)

    def to_exiv2(self, tag):
        return {'Exif': self.to_exif,
                'Iptc': self.to_iptc,
                'Xmp': self.to_xmp}[tag.split('.')[0]]()

    def to_exif(self):
        return str(self)

    def to_iptc(self):
        return str(self)

    def to_xmp(self):
        return str(self)

    def compact_form(self):
        return self

    def merge(self, info, tag, other):
        result, merged, ignored = self.merge_item(self, other)
        if ignored:
            self.log_ignored(info, tag, other)
        elif merged:
            self.log_merged(info, tag, other)
            return self.__class__(result)
        return self

    def merge_item(self, this, other):
        if self.contains(this, other):
            return this, False, False
        if self.contains(other, this):
            return other, True, False
        return self.concat(this, other)

    def contains(self, this, other):
        return other == this

    def concat(self, this, other):
        return this, False, True

    @staticmethod
    def log_merged(info, tag, value):
        logger.info('%s: merged %s', info, tag)

    def log_replaced(self, info, tag, value):
        logger.log(
            (logging.WARNING, logging.INFO)[self._quiet],
            '%s: "%s" replaced by %s "%s"', info, str(self), tag, str(value))

    @classmethod
    def log_ignored(cls, info, tag, value):
        logger.log(
            (logging.WARNING, logging.INFO)[cls._quiet],
            '%s: ignored %s "%s"', info, tag, str(value))


class MD_UnmergableString(MD_Value, str):
    def __new__(cls, value=None):
        if value is None:
            value = ''
        elif isinstance(value, str):
            value = value.strip()
        return super(MD_UnmergableString, cls).__new__(cls, value)

    @classmethod
    def from_exiv2(cls, file_value, tag):
        if isinstance(file_value, list):
            file_value = ' // '.join(file_value)
        return cls(file_value)

    def contains(self, this, other):
        return other in this


class MD_String(MD_UnmergableString):
    def concat(self, this, other):
        return this + ' // ' + other, True, False


class MD_Software(MD_String):
    @classmethod
    def from_exiv2(cls, file_value, tag):
        if tag.startswith('Iptc'):
            file_value = ' v'.join(x for x in file_value if x)
        return cls(file_value)

    def to_iptc(self):
        return self.split(' v')


class MD_Dict(MD_Value, dict):
    def __init__(self, value=None):
        value = value or {}
        # can initialise from a string containing comma separated values
        if isinstance(value, str):
            value = value.split(',')
        # or a list of values
        if isinstance(value, (tuple, list)):
            value = zip(self._keys, value)
        # initialise all keys to None
        result = dict.fromkeys(self._keys)
        # update with any supplied values
        if value:
            result.update(value)
        # let sub-classes do any data manipulation
        result = self.convert(result)
        super(MD_Dict, self).__init__(result)

    @staticmethod
    def convert(value):
        for key in value:
            if isinstance(value[key], str):
                value[key] = value[key].strip() or None
        return value

    def __setattr__(self, name, value):
        raise TypeError(
            "{} does not support item assignment".format(self.__class__))

    def __setitem__(self, key, value):
        raise TypeError(
            "{} does not support item assignment".format(self.__class__))

    def __bool__(self):
        return any([x is not None for x in self.values()])

    def to_exif(self):
        return [self[x] for x in self._keys]

    def __str__(self):
        return '\n'.join('{}: {}'.format(k, v) for (k, v) in self.items() if v)


class MD_DateTime(MD_Dict):
    # store date and time with "precision" to store how much is valid
    # tz_offset is stored in minutes
    _keys = ('datetime', 'precision', 'tz_offset')

    @classmethod
    def convert(cls, value):
        value['precision'] = value['precision'] or 7
        if value['datetime']:
            value['datetime'] = cls.truncate_datetime(
                value['datetime'], value['precision'])
        if value['precision'] <= 3:
            value['tz_offset'] = None
        return value

    _replace = (('microsecond', 0), ('second', 0),
                ('minute',      0), ('hour',   0),
                ('day',         1), ('month',  1))

    @classmethod
    def truncate_datetime(cls, date_time, precision):
        return date_time.replace(**dict(cls._replace[:7 - precision]))

    _tz_re = re.compile(r'(.*?[T ].*?)([+-])(\d{1,2}):?(\d{1,2})$')
    _subsec_re = re.compile(r'(.*?)\.(\d+)$')
    _time_re = re.compile(r'(.*?)[T ](\d{1,2}):?(\d{1,2})?:?(\d{1,2})?$')
    _date_re = re.compile(r'(\d{1,4})[:-]?(\d{1,2})?[:-]?(\d{1,2})?$')

    @classmethod
    def from_ISO_8601(cls, datetime_string, sub_sec_string=None):
        """Sufficiently general ISO 8601 parser.

        See https://en.wikipedia.org/wiki/ISO_8601

        """
        if not datetime_string:
            return cls([])
        unparsed = datetime_string
        precision = 7
        # extract time zone
        match = cls._tz_re.match(unparsed)
        if match:
            unparsed, sign, hours, minutes = match.groups()
            tz_offset = (int(hours) * 60) + int(minutes)
            if sign == '-':
                tz_offset = -tz_offset
        elif unparsed[-1] == 'Z':
            tz_offset = 0
            unparsed = unparsed[:-1]
        else:
            tz_offset = None
        # extract sub seconds
        if not sub_sec_string:
            match = cls._subsec_re.match(unparsed)
            if match:
                unparsed, sub_sec_string = match.groups()
        if sub_sec_string:
            sub_sec_string = sub_sec_string.strip()
            microsecond = int((sub_sec_string + '000000')[:6])
        else:
            microsecond = 0
            precision = 6
        # extract time
        match = cls._time_re.match(unparsed)
        if match:
            groups = match.groups('0')
            unparsed = groups[0]
            hour, minute, second = [int(x) for x in groups[1:]]
            if match.lastindex < 4:
                precision = 2 + match.lastindex
        else:
            hour, minute, second = 0, 0, 0
            precision = 3
        # extract date
        match = cls._date_re.match(unparsed)
        if match:
            year, month, day = [int(x) for x in match.groups('1')]
            if match.lastindex < 3:
                precision = match.lastindex
            if day == 0:
                day = 1
                precision = 2
            if month == 0:
                month = 1
                precision = 1
        else:
            raise ValueError(
                'Cannot parse datetime "{}"'.format(datetime_string))
        return cls((
            datetime(year, month, day, hour, minute, second, microsecond),
            precision, tz_offset))

    _fmt_elements = ('%Y', '-%m', '-%d', 'T%H', ':%M', ':%S', '.%f')

    def to_ISO_8601(self, precision=None, time_zone=True):
        if precision is None:
            precision = self['precision']
        fmt = ''.join(self._fmt_elements[:precision])
        datetime_string = self['datetime'].strftime(fmt)
        if precision > 6 and datetime_string[-3:] == '000':
            # truncate subsecond to 3 digits
            datetime_string = datetime_string[:-3]
        if precision > 3 and time_zone and self['tz_offset'] is not None:
            # add time zone
            minutes = self['tz_offset']
            if minutes >= 0:
                datetime_string += '+'
            else:
                datetime_string += '-'
                minutes = -minutes
            datetime_string += '{:02d}:{:02d}'.format(
                minutes // 60, minutes % 60)
        return datetime_string

    @classmethod
    def from_ffmpeg(cls, file_value, tag):
        return cls.from_ISO_8601(file_value)

    # many quicktime movies use Apple's 1904 timestamp zero point
    _qt_offset = (datetime(1970, 1, 1) - datetime(1904, 1, 1)).total_seconds()

    @classmethod
    def from_exiv2(cls, file_value, tag):
        if tag.startswith('Exif'):
            return cls.from_exif(file_value)
        if tag.startswith('Iptc'):
            return cls.from_iptc(file_value)
        if tag.startswith('Xmp.video'):
            try:
                time_stamp = int(file_value)
            except Exception:
                # not an integer timestamp
                return cls([])
            if not time_stamp:
                return cls([])
            # assume date should be in range 1970 to 2034
            if time_stamp > cls._qt_offset:
                time_stamp -= cls._qt_offset
            return cls((datetime.utcfromtimestamp(time_stamp), 6, None))
        return cls.from_ISO_8601(file_value)

    # From the Exif spec: "The format is "YYYY:MM:DD HH:MM:SS" with time
    # shown in 24-hour format, and the date and time separated by one
    # blank character [20.H]. When the date and time are unknown, all
    # the character spaces except colons (":") may be filled with blank
    # characters.

    # Although the standard says "all", I've seen examples where some of
    # the values are spaces, e.g. "2004:01:     :  :  ". I assume this
    # represents a reduced precision. In Photini we write a full
    # resolution datetime and get the precision from the Xmp value.
    @classmethod
    def from_exif(cls, file_value):
        datetime_string, sub_sec_string = file_value
        if not datetime_string:
            return cls([])
        # check for blank values
        while datetime_string[-2:] == '  ':
            datetime_string = datetime_string[:-3]
        # do conversion
        return cls.from_ISO_8601(datetime_string, sub_sec_string=sub_sec_string)

    def to_exif(self):
        datetime_string = self.to_ISO_8601(
            precision=max(self['precision'], 6), time_zone=False)
        date_string = datetime_string[:10].replace('-', ':')
        time_string = datetime_string[11:19]
        sub_sec_string = datetime_string[20:]
        return date_string + ' ' + time_string, sub_sec_string

    # The exiv2 library parses correctly formatted IPTC date & time and
    # gives us integer values for each element. If the date or time is
    # malformed we get a string instead, and ignore it.

    # The date (and time?) can have missing values represented by 00
    # according to
    # https://www.iptc.org/std/photometadata/specification/IPTC-PhotoMetadata#date-created
    @classmethod
    def from_iptc(cls, file_value):
        date_value, time_value = file_value
        if not date_value:
            return cls([])
        if isinstance(date_value, str):
            # Exiv2 couldn't read malformed date, let our parser have a go
            if isinstance(time_value, str):
                date_value += 'T' + time_value
            return cls.from_ISO_8601(date_value)
        if date_value['year'] == 0:
            return cls([])
        precision = 3
        if isinstance(time_value, dict):
            tz_offset = (time_value['tzHour'] * 60) + time_value['tzMinute']
            del time_value['tzHour'], time_value['tzMinute']
            # all-zero time is assumed to be no time info
            if any(time_value.values()):
                precision = 6
        else:
            # missing or malformed time
            time_value = {}
            tz_offset = None
        if date_value['day'] == 0:
            date_value['day'] = 1
            precision = 2
        if date_value['month'] == 0:
            date_value['month'] = 1
            precision = 1
        return cls((datetime(**date_value, **time_value), precision, tz_offset))

    def to_iptc(self):
        precision = self['precision']
        datetime = self['datetime']
        year, month, day = datetime.year, datetime.month, datetime.day
        if precision < 2:
            month = 0
        if precision < 3:
            day = 0
        date_value = year, month, day
        if precision < 4:
            time_value = None
        else:
            tz_offset = self['tz_offset']
            if tz_offset is None:
                tz_hr, tz_min = 0, 0
            else:
                tz_hr, tz_min = tz_offset // 60, tz_offset % 60
            time_value = (
                datetime.hour, datetime.minute, datetime.second, tz_hr, tz_min)
        return date_value, time_value

    # XMP uses extended ISO 8601, but the time cannot be hours only. See
    # p75 of
    # https://partners.adobe.com/public/developer/en/xmp/sdk/XMPspecification.pdf
    # According to p71, when converting Exif values with no time zone,
    # local time zone should be assumed. However, the MWG guidelines say
    # this must not be assumed to be the time zone where the photo is
    # processed. It also says the XMP standard has been revised to make
    # time zone information optional.
    def to_xmp(self):
        precision = self['precision']
        if precision == 4:
            precision = 5
        return self.to_ISO_8601(precision=precision)

    def __bool__(self):
        return bool(self['datetime'])

    def __str__(self):
        return self.to_ISO_8601()

    def to_utc(self):
        if self['tz_offset']:
            return self['datetime'] - timedelta(minutes=self['tz_offset'])
        return self['datetime']

    def merge(self, info, tag, other):
        if other == self or not other:
            return self
        if other['datetime'] != self['datetime']:
            verbose = (other['datetime'] != self.truncate_datetime(
                self['datetime'], other['precision']))
            # datetime values differ, choose self or other
            if (self['tz_offset'] in (None, 0)) != (other['tz_offset'] in (None, 0)):
                if self['tz_offset'] in (None, 0):
                    # other has "better" time zone info so choose it
                    if verbose:
                        self.log_replaced(info, tag, other)
                    return other
                # self has better time zone info
                if verbose:
                    self.log_ignored(info, tag, other)
                return self
            if other['precision'] > self['precision']:
                # other has higher precision so choose it
                if verbose:
                    self.log_replaced(info, tag, other)
                return other
            if verbose:
                self.log_ignored(info, tag, other)
            return self
        # datetime values agree, merge other info
        result = dict(self)
        if tag.startswith('Xmp'):
            # other is Xmp, so has trusted timezone and precision
            result['precision'] = other['precision']
            result['tz_offset'] = other['tz_offset']
        else:
            # use higher precision
            if other['precision'] > self['precision']:
                result['precision'] = other['precision']
            # only trust non-zero timezone (IPTC defaults to zero)
            if (self['tz_offset'] in (None, 0)
                    and other['tz_offset'] not in (None, 0)):
                result['tz_offset'] = other['tz_offset']
        return MD_DateTime(result)


class MD_LensSpec(MD_Dict):
    # simple class to store lens "specificaton"
    _keys = ('min_fl', 'max_fl', 'min_fl_fn', 'max_fl_fn')
    _quiet = True

    @staticmethod
    def convert(value):
        for key in value:
            value[key] = safe_fraction(value[key] or 0)
        return value

    @classmethod
    def from_exiv2(cls, file_value, tag):
        if not file_value:
            return cls([])
        if isinstance(file_value, str):
            file_value = file_value.split()
        if 'CanonCs' in tag:
            long_focal, short_focal, focal_units = [int(x) for x in file_value]
            if focal_units == 0:
                return cls([])
            file_value = [(short_focal, focal_units), (long_focal, focal_units)]
        return cls(file_value)

    def to_xmp(self):
        return ' '.join(['{}/{}'.format(x.numerator, x.denominator)
                         for x in self.to_exif()])

    def __str__(self):
        return ','.join(['{:g}'.format(float(self[x])) for x in self._keys])


class MD_Thumbnail(MD_Dict):
    _keys = ('w', 'h', 'fmt', 'data', 'image')
    _quiet = True

    @staticmethod
    def image_from_data(data):
        # PyQt5 seems to be the only thing that can use memoryviews
        if isinstance(data, memoryview) and (
                using_pyside or qt_version_info >= (6, 0)):
            data = bytes(data)
        buf = QtCore.QBuffer()
        buf.setData(data)
        reader = QtGui.QImageReader(buf)
        fmt = reader.format().data().decode().upper()
        reader.setAutoTransform(False)
        image = reader.read()
        if image.isNull():
            raise RuntimeError(reader.errorString())
        return fmt, image

    @staticmethod
    def data_from_image(image, max_size=60000):
        buf = QtCore.QBuffer()
        buf.open(buf.OpenModeFlag.WriteOnly)
        quality = 95
        while quality > 10:
            image.save(buf, 'JPEG', quality)
            data = buf.data().data()
            if len(data) < max_size:
                return data
            quality -= 5
        return None

    @classmethod
    def convert(cls, value):
        value['fmt'] = value['fmt'] or 'JPEG'
        if value['data'] and not value['image']:
            value['fmt'], value['image'] = cls.image_from_data(value['data'])
        if not value['image']:
            return {}
        value['w'] = value['image'].width()
        value['h'] = value['image'].height()
        if value['data'] and len(value['data']) >= 60000:
            # don't keep unusably large amount of data
            value['data'] = None
        return value

    def to_exif(self):
        fmt, data = self['fmt'], self['data']
        if not data:
            fmt = 'JPEG'
            data = self.data_from_image(self['image'])
        if not data:
            return None, None, None, None
        fmt = (None, 6)[fmt == 'JPEG']
        return self['w'], self['h'], fmt, data

    def to_xmp(self):
        fmt, data = self['fmt'], self['data']
        if fmt != 'JPEG':
            data = None
        if not data:
            fmt = 'JPEG'
            data = self.data_from_image(self['image'], max_size=2**32)
        data = codecs.encode(data, 'base64_codec').decode('ascii')
        return [{
            'xmpGImg:width': str(self['w']),
            'xmpGImg:height': str(self['h']),
            'xmpGImg:format': fmt,
            'xmpGImg:image': data,
            }]

    def __str__(self):
        result = '{fmt} thumbnail, {w}x{h}'.format(**self)
        if self['data']:
            result += ', {} bytes'.format(len(self['data']))
        return result


class MD_Collection(MD_Dict):
    # class for a group of independent items, each of which is an MD_Value
    _type = {}
    _default_type = MD_String

    @classmethod
    def get_type(cls, key):
        if key in cls._type:
            return cls._type[key]
        return cls._default_type

    @classmethod
    def convert(cls, value):
        for key in value:
            if not value[key]:
                continue
            value[key] = cls.get_type(key)(value[key])
        return value

    @classmethod
    def from_exiv2(cls, file_value, tag):
        if not (file_value and any(file_value)):
            return cls([])
        value = dict(zip(cls._keys, file_value))
        for key in value:
            value[key] = cls.get_type(key).from_exiv2(value[key], tag)
        return cls(value)

    def to_exif(self):
        return [(self[x] or None) and self[x].to_exif() for x in self._keys]

    def to_iptc(self):
        return [(self[x] or None) and self[x].to_iptc() for x in self._keys]

    def to_xmp(self):
        return [(self[x] or None) and self[x].to_xmp() for x in self._keys]

    def merge(self, info, tag, other):
        if other == self:
            return self
        result = dict(self)
        for key in other:
            if other[key] is None:
                continue
            if key in result and result[key] is not None:
                result[key], merged, ignored = result[key].merge_item(
                                                        result[key], other[key])
            else:
                result[key] = other[key]
                merged, ignored = True, False
            if ignored:
                self.log_ignored(info, tag, {key: str(other[key])})
            elif merged:
                self.log_merged(info, tag, {key: str(other[key])})
        return self.__class__(result)


class MD_Structure(MD_Value, dict):
    extendable = False

    def __init__(self, value=None):
        value = value or {}
        # deep copy initial values
        value = dict((k, self.get_type(k, v)(v)) for (k, v) in value.items())
        # set missing values to empty
        for k in self.item_type:
            if k not in value:
                value[k] = self.item_type[k]()
        super(MD_Structure, self).__init__(value)

    @classmethod
    def get_type(cls, key, value):
        if cls.extendable and key not in cls.item_type:
            logger.warning('Inferring type for %s', key)
            if isinstance(value, (list, tuple)):
                cls.item_type[key] = MD_MultiString
            elif isinstance(value, dict):
                cls.item_type[key] = MD_LangAlt
            else:
                cls.item_type[key] = MD_String
        return cls.item_type[key]

    @classmethod
    def from_exiv2(cls, file_value, tag):
        file_value = file_value or {}
        if isinstance(file_value, (list, tuple)):
            # "legacy" list of string values
            file_value = dict(zip(cls.legacy_keys, file_value))
        for key, value in file_value.items():
            file_value[key] = cls.get_type(key, value).from_exiv2(value, tag)
        return cls(file_value)

    def merge(self, info, tag, other):
        if other == self:
            return self
        result = dict(self)
        for key in other:
            if not other[key]:
                continue
            if key in result and result[key]:
                result[key] = result[key].merge(
                    info, '{}[{}]'.format(tag, key), other[key])
            else:
                result[key] = other[key]
                self.log_merged(info, '{}[{}]'.format(tag, key), other[key])
        return self.__class__(result)

    def to_exif(self):
        if not self:
            return None
        return [self[k] and self[k].to_exif() for k in self.legacy_keys]

    def to_iptc(self):
        if not self:
            return None
        return [self[k] and self[k].to_iptc() for k in self.legacy_keys]

    def to_xmp(self):
        if not self:
            return None
        return dict((k, v.to_xmp()) for (k, v) in self.items() if v)

    def compact_form(self):
        return dict((k.split(':')[-1], v.compact_form())
                    for (k, v) in self.items() if v)

    def __str__(self):
        return pprint.pformat(self.compact_form(), compact=True)

    def __bool__(self):
        return any(self.values())


class Unused(object):
    def __new__(cls, value=None):
        return None

    @classmethod
    def from_exiv2(cls, file_value, tag):
        logger.warning('%s: to be deleted when data is saved: %s',
                       tag, file_value)
        return None


class MD_ContactInformation(MD_Structure):
    item_type = {
        'plus:LicensorID': Unused,
        'plus:LicensorName': Unused,
        'plus:LicensorStreetAddress': MD_String,
        'plus:LicensorExtendedAddress': MD_String,
        'plus:LicensorCity': MD_String,
        'plus:LicensorRegion': MD_String,
        'plus:LicensorPostalCode': MD_String,
        'plus:LicensorCountry': MD_String,
        'plus:LicensorTelephoneType1': Unused,
        'plus:LicensorTelephone1': MD_String,
        'plus:LicensorTelephoneType2': Unused,
        'plus:LicensorTelephone2': Unused,
        'plus:LicensorEmail': MD_String,
        'plus:LicensorURL': MD_String,
        }

    _ci_map = {
        'Iptc4xmpCore:CiAdrExtadr': 'plus:LicensorStreetAddress',
        'Iptc4xmpCore:CiAdrCity':   'plus:LicensorCity',
        'Iptc4xmpCore:CiAdrCtry':   'plus:LicensorCountry',
        'Iptc4xmpCore:CiEmailWork': 'plus:LicensorEmail',
        'Iptc4xmpCore:CiTelWork':   'plus:LicensorTelephone1',
        'Iptc4xmpCore:CiAdrPcode':  'plus:LicensorPostalCode',
        'Iptc4xmpCore:CiAdrRegion': 'plus:LicensorRegion',
        'Iptc4xmpCore:CiUrlWork':   'plus:LicensorURL',
        }

    @classmethod
    def from_exiv2(cls, file_value, tag):
        if tag == 'Xmp.iptc.CreatorContactInfo':
            file_value = file_value or {}
            file_value = dict((cls._ci_map[k], v)
                              for (k, v) in file_value.items())
            if 'plus:LicensorStreetAddress' in file_value:
                line1, sep, line2 = file_value[
                    'plus:LicensorStreetAddress'].partition('\n')
                if line2:
                    file_value['plus:LicensorExtendedAddress'] = line1
                    file_value['plus:LicensorStreetAddress'] = line2
        elif file_value:
            for value in file_value[1:]:
                logger.warning(
                    '%s: to be deleted when data is saved: %s', tag, value)
            # Xmp.plus.Licensor is an XMP bag with up to 3 entries, use the 1st
            file_value = file_value[0]
        return super(MD_ContactInformation, cls).from_exiv2(file_value, tag)

    def to_xmp(self):
        return [super(MD_ContactInformation, self).to_xmp()]


class MD_StructArray(MD_Value, tuple):
    # class for arrays of XMP structures such as locations or image regions
    def __new__(cls, value=None):
        value = value or []
        temp = []
        for item in value:
            if not isinstance(item, cls.item_type):
                item = cls.item_type(item)
            temp.append(item)
        while temp and not temp[-1]:
            temp = temp[:-1]
        return super(MD_StructArray, cls).__new__(cls, temp)

    @classmethod
    def from_exiv2(cls, file_value, tag):
        if not file_value:
            return cls()
        # Exif and IPTC only store one item, XMP stores any number
        if tag.startswith('Xmp'):
            file_value = [cls.item_type.from_exiv2(x, tag) for x in file_value]
        else:
            file_value = [cls.item_type.from_exiv2(file_value, tag)]
        return cls(file_value)

    def to_exif(self):
        return self and self[0].to_exif()

    def to_iptc(self):
        return self and self[0].to_iptc()

    def to_xmp(self):
        return [x.to_xmp() for x in self]

    def merge(self, info, tag, other):
        result = self
        for item in other:
            if not isinstance(item, self.item_type):
                item = self.item_type(item)
            idx = result.index(item)
            result = list(result)
            if idx < len(result):
                result[idx] = result[idx].merge(info, tag, item)
            else:
                self.log_merged(info, tag, item)
                result.append(item)
            result = self.__class__(result)
        return result

    def compact_form(self):
        return [v.compact_form() for v in self if v]

    def __str__(self):
        return '\n\n'.join(str(x) for x in self)


class MD_LangAlt(MD_Value, dict):
    # XMP LangAlt values are a sequence of RFC3066 language tag keys and
    # text values. The sequence can have a single default value, but if
    # it has more than one value, the default should be repeated with a
    # language tag. See
    # https://developer.adobe.com/xmp/docs/XMPNamespaces/XMPDataTypes/#language-alternative

    DEFAULT = 'x-default'

    def __init__(self, value=None, default_lang=None, strip=True):
        if isinstance(value, str):
            value = {self.DEFAULT: value}
        elif isinstance(value, MD_LangAlt):
            default_lang = default_lang or value.default_lang
        value = value or {}
        if strip:
            value = dict((k, v.strip()) for (k, v) in value.items())
        value = dict((k, v) for (k, v) in value.items() if v)
        if default_lang:
            self.default_lang = default_lang
            if self.DEFAULT in value and self.default_lang not in value:
                value[self.default_lang] = value[self.DEFAULT]
        else:
            self.default_lang = self.identify_default(value)
        if self.default_lang != self.DEFAULT and self.DEFAULT in value:
            del value[self.DEFAULT]
        super(MD_LangAlt, self).__init__(value)

    @classmethod
    def identify_default(cls, value):
        keys = list(value.keys())
        if len(keys) == 0:
            return QtCore.QLocale.system().bcp47Name() or cls.DEFAULT
        if len(keys) == 1:
            return keys[0]
        if cls.DEFAULT not in keys:
            return cls.DEFAULT
        # look for language with same text as 'x-default' value
        text = value[cls.DEFAULT]
        for k, v in value.items():
            if k != cls.DEFAULT and v == text:
                return k
        return cls.DEFAULT

    def find_key(self, key):
        # languages are case insensitive
        key = key.lower()
        for k in self:
            if k.lower() == key:
                return k
        return None

    def __contains__(self, key):
        key = key and self.find_key(key)
        return super(MD_LangAlt, self).__contains__(key)

    def __getitem__(self, key):
        key = self.find_key(key)
        if not key:
            return ''
        return super(MD_LangAlt, self).__getitem__(key)

    def __setitem__(self, key, value):
        raise NotImplemented()

    def __bool__(self):
        return any(self.values())

    def __eq__(self, other):
        if isinstance(other, MD_LangAlt):
            return not self.__ne__(other)
        return super(MD_LangAlt, self).__eq__(other)

    def __ne__(self, other):
        if isinstance(other, MD_LangAlt):
            if self.default_lang != other.default_lang:
                return True
        return super(MD_LangAlt, self).__ne__(other)

    def __str__(self):
        result = []
        for key in self.languages():
            if key != self.DEFAULT:
                result.append('-- {} --'.format(key))
            result.append(self[key])
        return '\n'.join(result)

    def best_match(self, lang=None):
        if len(self) < 2:
            return self[self.default_lang]
        lang = lang or QtCore.QLocale.system().bcp47Name()
        if not lang:
            return self[self.default_lang]
        lang = lang.lower()
        langs = [lang]
        if '-' in lang:
            langs.append(lang.split('-')[0])
        for lang in langs:
            for k in self:
                if k.lower() == lang:
                    return self[k]
            for k in self:
                if k.lower().startswith(lang):
                    return self[k]
        return self[self.default_lang]

    def languages(self):
        keys = list(self.keys())
        if len(keys) < 1:
            return [self.default_lang]
        if len(keys) < 2:
            return keys
        if self.default_lang in keys:
            keys.remove(self.default_lang)
            keys.sort(key=lambda x: x.lower())
            keys.insert(0, self.default_lang)
        else:
            keys.sort(key=lambda x: x.lower())
        return keys

    def to_exif(self):
        # Xmp spec says to store only the default language in Exif
        if not self:
            return None
        return self[self.default_lang]

    def to_iptc(self):
        return self.to_exif()

    def to_xmp(self):
        if not self:
            return None
        if len(self) < 2:
            return dict(self)
        result = dict((k, v) for (k, v) in self.items() if v)
        if len(result) > 1:
            result[self.DEFAULT] = result[self.default_lang]
        return result

    def merge(self, info, tag, other):
        if self == other:
            return self
        if not isinstance(other, MD_LangAlt):
            other = MD_LangAlt(other)
        default_lang = self.default_lang
        if default_lang == self.DEFAULT:
            default_lang = other.default_lang
        result = dict(self)
        for key, value in other.items():
            if key == self.DEFAULT:
                # try to find matching value
                for k, v in result.items():
                    if k != self.DEFAULT and v == value:
                        key = k
                        break
            else:
                # try to find matching language
                key = self.find_key(key) or key
            if key not in result:
                result[key] = value
            elif value in result[key]:
                continue
            elif result[key] in value:
                result[key] = value
            else:
                result[key] += ' // ' + value
            self.log_merged(info + '[' + key + ']', tag, value)
        return self.__class__(result, default_lang=default_lang)


class MD_Rights(MD_Collection):
    # stores IPTC rights information
    _keys = ('UsageTerms', 'WebStatement')
    _default_type = MD_UnmergableString
    _type = {'UsageTerms': MD_LangAlt}


class MD_CameraModel(MD_Collection):
    _keys = ('make', 'model', 'serial_no')
    _default_type = MD_UnmergableString
    _quiet = True

    def convert(self, value):
        if value['model'] == 'unknown':
            value['model'] = None
        return super(MD_CameraModel, self).convert(value)

    def __str__(self):
        return str(dict([(x, y) for x, y in self.items() if y]))

    def get_name(self, inc_serial=True):
        result = []
        # start with 'model'
        if self['model']:
            result.append(self['model'])
        # only add 'make' if it's not part of model
        if self['make']:
            if not (result
                    and self['make'].split()[0].lower() in result[0].lower()):
                result = [self['make']] + result
        # add serial no if a unique answer is needed
        if inc_serial and self['serial_no']:
            result.append('(S/N: ' + self['serial_no'] + ')')
        return ' '.join(result)


class MD_LensModel(MD_Collection):
    _keys = ('make', 'model', 'serial_no', 'spec')
    _default_type = MD_UnmergableString
    _type = {'spec': MD_LensSpec}
    _quiet = True

    def convert(self, value):
        if value['model'] in ('n/a', '(0)', '65535'):
            value['model'] = None
        if value['serial_no'] == '0000000000':
            value['serial_no'] = None
        return super(MD_LensModel, self).convert(value)

    def get_name(self, inc_serial=True):
        result = []
        # start with 'model'
        if self['model']:
            result.append(self['model'])
        # only add 'make' if it's not part of model
        if self['make']:
            if not (result
                    and self['make'].split()[0].lower() in result[0].lower()):
                result = [self['make']] + result
        if inc_serial and self['serial_no']:
            result.append('(S/N: ' + self['serial_no'] + ')')
        if self['spec'] and not result:
            # generic name based on spec
            fl = [float(self['spec']['min_fl']), float(self['spec']['max_fl'])]
            fl = '–'.join(['{:g}'.format(x) for x in fl if x])
            fn = [float(self['spec']['min_fl_fn']),
                  float(self['spec']['max_fl_fn'])]
            fn = '–'.join(['{:g}'.format(x) for x in fn if x])
            if fl:
                model = fl + ' mm'
                if fn:
                    model += ' ƒ/' + fn
                result.append(model)
        return ' '.join(result)


class MD_MultiString(MD_Value, tuple):
    def __new__(cls, value=None):
        value = value or []
        if isinstance(value, str):
            value = value.split(';')
        value = filter(bool, [x.strip() for x in value])
        return super(MD_MultiString, cls).__new__(cls, value)

    def to_exif(self):
        return ';'.join(self)

    def to_iptc(self):
        return tuple(self)

    def to_xmp(self):
        return tuple(self)

    def __str__(self):
        return '; '.join(self)

    def merge(self, info, tag, other):
        merged = False
        result = list(self)
        for item in other:
            if tag.split('.')[0] == 'Iptc':
                # IPTC-IIM data can be truncated version of existing value
                if item not in [MetadataHandler.truncate_iptc(tag, x)
                                for x in result]:
                    result.append(item)
                    merged = True
            elif item not in result:
                result.append(item)
                merged = True
        if merged:
            self.log_merged(info, tag, other)
            return MD_MultiString(result)
        return self


class MD_Keywords(MD_MultiString):
    _machine_tag = re.compile(r'^(.+):(.+)=(.+)$')

    def human_tags(self):
        return [x for x in self if not self._machine_tag.match(x)]

    def machine_tags(self):
        # yield keyword, (ns, predicate, value) for each machine tag
        for keyword in self:
            match = self._machine_tag.match(keyword)
            if match:
                yield keyword, match.groups()


class MD_Int(MD_Value, int):
    def __new__(cls, value=None):
        if value is None:
            return None
        return super(MD_Int, cls).__new__(cls, value)

    def to_exif(self):
        return self

    def __bool__(self):
        # reinterpret to mean "has a value", even if the value is zero
        return True


class MD_Orientation(MD_Int):
    @classmethod
    def from_ffmpeg(cls, file_value, tag):
        mapping = {'0': 1, '90': 6, '180': 3, '-90': 8}
        if file_value not in mapping:
            raise ValueError('unrecognised orientation {}'.format(file_value))
        return cls(mapping[file_value])

    def get_transform(self):
        bits = self - 1
        if not bits:
            return None
        # need to rotate and or reflect image
        # translation is set so a unit rectangle maps to a unit rectangle
        transform = QtGui.QTransform()
        if bits & 0b001:
            # reflect left-right
            transform = transform.scale(-1.0, 1.0)
        if bits & 0b010:
            # rotate 180°
            transform = transform.rotate(180.0)
        if bits & 0b100:
            # rotate 90° then reflect left-right
            transform = transform.rotate(-90.0)
            transform = transform.scale(-1.0, 1.0)
        if transform.m11() + transform.m12() < 0:
            transform = transform.translate(-1, 0)
        if transform.m21() + transform.m22() < 0:
            transform = transform.translate(0, -1)
        return transform


class MD_Timezone(MD_Int):
    _quiet = True

    @classmethod
    def from_exiv2(cls, file_value, tag):
        if file_value is None:
            return None
        if tag == 'Exif.Image.TimeZoneOffset':
            # convert hours to minutes
            file_value = file_value * 60
        return cls(file_value)


class MD_Float(MD_Value, float):
    def __new__(cls, value=None):
        if value is None:
            return None
        return super(MD_Float, cls).__new__(cls, value)

    def __bool__(self):
        # reinterpret to mean "has a value", even if the value is zero
        return True


class MD_Rating(MD_Float):
    @classmethod
    def from_exiv2(cls, file_value, tag):
        if not file_value:
            return None
        if tag in ('Exif.Image.RatingPercent', 'Xmp.MicrosoftPhoto.Rating'):
            value = 1.0 + (float(file_value) / 25.0)
        else:
            value = min(max(float(file_value), -1.0), 5.0)
        return cls(value)

    def to_exif(self):
        return str(int(self + 1.5) - 1)


class MD_Rational(MD_Value, Fraction):
    def __new__(cls, value=None):
        if value is None:
            return None
        return super(MD_Rational, cls).__new__(cls, safe_fraction(value))

    def to_exif(self):
        return self

    def to_xmp(self):
        return '{}/{}'.format(self.numerator, self.denominator)

    def compact_form(self):
        return float(self)

    def __bool__(self):
        # reinterpret to mean "has a value", even if the value is zero
        return True

    def __str__(self):
        return str(float(self))


class MD_Altitude(MD_Rational):
    @classmethod
    def from_exiv2(cls, file_value, tag):
        if not all(file_value):
            return None
        altitude, ref = file_value
        altitude = safe_fraction(altitude)
        if ref in (b'\x01', '1'):
            altitude = -altitude
        return cls(altitude)

    def to_exif(self):
        altitude = self
        if altitude < 0:
            altitude = -altitude
            ref = b'\x01'
        else:
            ref = b'\x00'
        return altitude, ref

    def to_xmp(self):
        altitude = self
        if altitude < 0:
            altitude = -altitude
            ref = '1'
        else:
            ref = '0'
        return '{}/{}'.format(altitude.numerator, altitude.denominator), ref

    def contains(self, this, other):
        return abs(float(other) - float(this)) < 0.001


class MD_Coordinate(MD_Rational):
    @classmethod
    def from_exiv2(cls, file_value, tag):
        if tag.startswith('Exif'):
            return cls.from_exif(file_value)
        return cls.from_xmp(file_value)

    @classmethod
    def from_exif(cls, value):
        if not all(value):
            return None
        value, ref = value
        value = [safe_fraction(x, limit=False) for x in value]
        degrees, minutes, seconds = value
        degrees += (minutes / 60) + (seconds / 3600)
        if ref in ('S', 'W'):
            degrees = -degrees
        return cls(degrees)

    @classmethod
    def from_xmp(cls, value):
        if not value:
            return None
        ref = value[-1]
        if ref in ('N', 'E', 'S', 'W'):
            negative = ref in ('S', 'W')
            value = value[:-1]
        else:
            logger.warning('no direction in XMP GPSCoordinate: %s', value)
            negative = False
        if value[0] in ('+', '-'):
            logger.warning(
                'incorrect use of signed XMP GPSCoordinate: %s', value)
            if value[0] == '-':
                negative = not negative
            value = value[1:]
        value = [safe_fraction(x, limit=False) for x in value.split(',')]
        degrees, minutes = value[:2]
        degrees += (minutes / 60)
        if len(value) > 2:
            seconds = value[2]
            degrees += (seconds / 3600)
        if negative:
            degrees = -degrees
        return cls(degrees)

    def to_exif_part(self):
        degrees = self
        pstv = degrees >= 0
        if not pstv:
            degrees = -degrees
        # make degrees and minutes integer (not mandated by Exif, but typical)
        i = int(degrees)
        minutes = (degrees - i) * 60
        degrees = Fraction(i)
        i = int(minutes)
        seconds = (minutes - i) * 60
        minutes = Fraction(i)
        seconds = seconds.limit_denominator(1000000)
        return (degrees, minutes, seconds), pstv

    def to_xmp_part(self):
        numbers, pstv = self.to_exif_part()
        if all([x.denominator == 1 for x in numbers]):
            return ('{:d},{:d},{:d}'.format(*[x.numerator for x in numbers]),
                    pstv)
        degrees, minutes, seconds = numbers
        degrees = int(degrees)
        minutes = float(minutes + (seconds / 60))
        return '{:d},{:.8f}'.format(degrees, minutes), pstv

    def contains(self, this, other):
        if other is None:
            return False
        return abs(float(other) - float(this)) < 0.0000005

    def compact_form(self):
        return round(float(self), 6)

    def __str__(self):
        return '{:.6f}'.format(float(self))

    def __eq__(self, other):
        return self.contains(self, other)

    def __ne__(self, other):
        return not self.contains(self, other)


class MD_Latitude(MD_Coordinate):
    def to_exif(self):
        numbers, pstv = self.to_exif_part()
        return numbers, ('S', 'N')[pstv]

    def to_xmp(self):
        string, pstv = self.to_xmp_part()
        return string + ('S', 'N')[pstv]


class MD_Longitude(MD_Coordinate):
    def to_exif(self):
        numbers, pstv = self.to_exif_part()
        return numbers, ('W', 'E')[pstv]

    def to_xmp(self):
        string, pstv = self.to_xmp_part()
        return string + ('W', 'E')[pstv]


class GPSVersionId(MD_Value, bytes):
    def __new__(cls, value=None):
        value = value or b'\x02\x00\x00\x00'
        return super(GPSVersionId, cls).__new__(cls, value)

    @classmethod
    def from_exiv2(cls, file_value, tag):
        if file_value and tag.startswith('Xmp'):
            file_value = [int(x) for x in file_value.split('.')]
        return cls(file_value)

    def to_exif(self):
        return self

    def to_xmp(self):
        return '.'.join(str(x) for x in self)

    def compact_form(self):
        return self.to_xmp()


class MD_GPSinfo(MD_Structure):
    item_type = {
        'version_id': GPSVersionId,
        'method': MD_UnmergableString,
        'exif:GPSAltitude': MD_Altitude,
        'exif:GPSLatitude': MD_Latitude,
        'exif:GPSLongitude': MD_Longitude,
        }
    legacy_keys = (
        'version_id', 'method',
        'exif:GPSAltitude', 'exif:GPSLatitude', 'exif:GPSLongitude')

    @classmethod
    def from_ffmpeg(cls, file_value, tag):
        if file_value:
            match = re.match(
                r'([-+]\d+\.\d+)([-+]\d+\.\d+)([-+]\d+\.\d+)?/', file_value)
            if match:
                return cls(dict(zip(('exif:GPSLatitude', 'exif:GPSLongitude',
                                     'exif:GPSAltitude'), match.groups())))
        return cls()

    @classmethod
    def from_exiv2(cls, file_value, tag):
        if tag.startswith('Xmp.video'):
            return cls.from_ffmpeg(file_value, tag)
        version_id = file_value[0]
        method = file_value[1]
        alt = file_value[2:4]
        if tag.startswith('Exif'):
            lat = file_value[4:6]
            lon = file_value[6:8]
        else:
            lat = file_value[4]
            lon = file_value[5]
        file_value = version_id, method, alt, lat, lon
        return super(MD_GPSinfo, cls).from_exiv2(file_value, tag)

    def to_exif(self):
        if not self:
            return None
        result = []
        for k in self.legacy_keys:
            if k in ('exif:GPSAltitude', 'exif:GPSLatitude',
                     'exif:GPSLongitude'):
                if self[k]:
                    result += self[k].to_exif()
                else:
                    result += [None, None]
            else:
                result.append(self[k] and self[k].to_exif())
        return result

    def to_xmp(self):
        if not self:
            return None
        result = []
        for k in self.legacy_keys:
            if k == 'exif:GPSAltitude':
                if self[k]:
                    result += self[k].to_xmp()
                else:
                    result += [None, None]
            else:
                result.append(self[k] and self[k].to_xmp())
        return result

    def __bool__(self):
        return any(self[k] for k in ('exif:GPSLatitude', 'exif:GPSLongitude',
                                     'exif:GPSAltitude'))

    def __eq__(self, other):
        return not self.__ne__(other)

    def __ne__(self, other):
        if not isinstance(other, MD_GPSinfo):
            other = MD_GPSinfo(other)
        return any(self[k] != other[k] for k in (
            'exif:GPSLatitude', 'exif:GPSLongitude', 'exif:GPSAltitude'))


class MD_Aperture(MD_Rational):
    # store FNumber and APEX aperture as fractions
    # only FNumber is presented to the user, either is computed if missing
    @classmethod
    def from_exiv2(cls, file_value, tag):
        if not any(file_value):
            return None
        f_number, apex = file_value
        if apex:
            apex = safe_fraction(apex)
        if not f_number:
            f_number = 2.0 ** (apex / 2.0)
        self = cls(f_number)
        if apex:
            self.apex = apex
        return self

    def to_exif(self):
        file_value = [self]
        if float(self) != 0:
            apex = getattr(self, 'apex', safe_fraction(math.log(self, 2) * 2.0))
            file_value.append(apex)
        return file_value

    def to_xmp(self):
        return ['{}/{}'.format(x.numerator, x.denominator)
                for x in self.to_exif()]

    def contains(self, this, other):
        return float(min(other, this)) > (float(max(other, this)) * 0.95)


class MD_VideoDuration(MD_Rational):
    @classmethod
    def from_ffmpeg(cls, file_value, tag):
        if tag == 'ffmpeg/streams[0]/duration':
            return cls(file_value)
        elif tag == 'ffmpeg/streams[0]/duration_ts':
            duration_ts, time_base = file_value
            if duration_ts and time_base:
                return cls(duration_ts * Fraction(time_base))
        elif tag == 'ffmpeg/streams[0]/frames':
            frames, frame_rate = file_value
            if frames and frame_rate:
                return cls((int(frames) / Fraction(frame_rate)))
        return None

    @classmethod
    def from_exiv2(cls, file_value, tag):
        if file_value:
            return cls((int(file_value), 1000))
        return None

    def contains(self, this, other):
        # some values are a bit approximate
        lo = float(min(other, this))
        hi = float(max(other, this))
        return hi - lo < max(hi * 0.0001, 0.2)


class MD_Dimensions(MD_Collection):
    _keys = ('width', 'height')
    _default_type = MD_Int

    def scaled_to(self, target_size):
        w = float(self['width'])
        h = float(self['height'])
        if w > h:
            return target_size, int((float(target_size) * h / w) + 0.5)
        return int((float(target_size) * w / h) + 0.5), target_size


class CountryCode(MD_UnmergableString):
    def __new__(cls, value=None):
        if value is None:
            value = ''
        elif isinstance(value, str):
            value = value.strip().upper()
        return super(CountryCode, cls).__new__(cls, value)


class MD_Location(MD_Structure):
    # stores IPTC defined location heirarchy
    item_type = {
        'Iptc4xmpExt:City': MD_String,
        'Iptc4xmpExt:CountryCode': CountryCode,
        'Iptc4xmpExt:CountryName': MD_String,
        'exif:GPSAltitude': MD_Rational,
        'exif:GPSLatitude': MD_Latitude,
        'exif:GPSLongitude': MD_Longitude,
        'Iptc4xmpExt:LocationId': MD_MultiString,
        'Iptc4xmpExt:LocationName': MD_LangAlt,
        'Iptc4xmpExt:ProvinceState': MD_String,
        'Iptc4xmpExt:Sublocation': MD_String,
        'Iptc4xmpExt:WorldRegion': MD_String,
        }
    legacy_keys = (
        'Iptc4xmpExt:Sublocation', 'Iptc4xmpExt:City',
        'Iptc4xmpExt:ProvinceState', 'Iptc4xmpExt:CountryName',
        'Iptc4xmpExt:CountryCode',
        )

    def to_xmp(self):
        if not self:
            # need a place holder for empty values
            return {'Iptc4xmpExt:City': ' '}
        return super(MD_Location, self).to_xmp()

    @classmethod
    def from_address(cls, gps, address, key_map):
        result = {}
        for key in cls.item_type:
            result[key] = []
        for key in key_map:
            for foreign_key in key_map[key]:
                if foreign_key not in address or not address[foreign_key]:
                    continue
                if key in result and address[foreign_key] not in result[key]:
                    result[key].append(address[foreign_key])
                del(address[foreign_key])
        # only use one country code
        result['Iptc4xmpExt:CountryCode'] = result[
            'Iptc4xmpExt:CountryCode'][:1]
        # put unknown foreign keys in Sublocation
        for foreign_key in address:
            if address[foreign_key] in ' '.join(
                    result['Iptc4xmpExt:Sublocation']):
                continue
            result['Iptc4xmpExt:Sublocation'] = [
                '{}: {}'.format(foreign_key, address[foreign_key])
                ] + result['Iptc4xmpExt:Sublocation']
        for key in result:
            result[key] = ', '.join(result[key]) or None
        result['exif:GPSLatitude'] = gps['lat']
        result['exif:GPSLongitude'] = gps['lng']
        return cls(result)


class MD_MultiLocation(MD_StructArray):
    item_type = MD_Location

    def index(self, other):
        for n, value in enumerate(self):
            if value == other:
                return n
        return len(self)


class MD_SingleLocation(MD_MultiLocation):
    def index(self, other):
        return 0


class ConceptIndentifier(MD_MultiString):
    def compact_form(self):
        fm = QtWidgets.QWidget().fontMetrics()
        width = fm.boundingRect('x' * 30).width()
        return [fm.elidedText(v, Qt.TextElideMode.ElideLeft, width)
                for v in self if v]


class EntityConcept(MD_Structure):
    item_type = {
        'Iptc4xmpExt:Name': MD_LangAlt,
        'xmp:Identifier': ConceptIndentifier,
        }


class EntityConceptArray(MD_StructArray):
    item_type = EntityConcept


class RegionBoundaryNumber(MD_Float):
    def set_decimals(self, decimals):
        self.decimals = decimals

    def compact_form(self):
        return round(self, self.decimals)

    def __str__(self):
        return '{:g}'.format(round(self, self.decimals))


class RegionBoundaryPoint(MD_Structure):
    item_type = {
        'Iptc4xmpExt:rbX': RegionBoundaryNumber,
        'Iptc4xmpExt:rbY': RegionBoundaryNumber,
        }

    def set_decimals(self, decimals):
        for v in self.values():
            v.set_decimals(decimals)


class RegionBoundaryPointArray(MD_StructArray):
    item_type = RegionBoundaryPoint

    def set_decimals(self, decimals):
        for v in self:
            v.set_decimals(decimals)


class RegionBoundary(MD_Structure):
    item_type = {
        'Iptc4xmpExt:rbShape': MD_String,
        'Iptc4xmpExt:rbUnit': MD_String,
        'Iptc4xmpExt:rbX': RegionBoundaryNumber,
        'Iptc4xmpExt:rbY': RegionBoundaryNumber,
        'Iptc4xmpExt:rbW': RegionBoundaryNumber,
        'Iptc4xmpExt:rbH': RegionBoundaryNumber,
        'Iptc4xmpExt:rbRx': RegionBoundaryNumber,
        'Iptc4xmpExt:rbVertices': RegionBoundaryPointArray,
        }

    def __init__(self, value=None):
        value = value or {}
        super(RegionBoundary, self).__init__(value)
        if self['Iptc4xmpExt:rbUnit'] == 'pixel':
            decimals = 0
        else:
            decimals = 4
        for v in self.values():
            if isinstance(v, (RegionBoundaryNumber, RegionBoundaryPointArray)):
                v.set_decimals(decimals)

    def compact_form(self):
        result = super(RegionBoundary, self).compact_form()
        if result['rbShape'] == 'circle':
            return {'rbShape':
                    'circle({rbX:g}, {rbY:g}, {rbRx:g})'.format(**result),
                    'rbUnit': result['rbUnit']}
        if result['rbShape'] == 'rectangle':
            return {'rbShape':
                    'rectangle({rbX:g}, {rbY:g}, {rbW:g}, {rbH:g})'.format(**result),
                    'rbUnit': result['rbUnit']}
        return result

    def to_Qt(self, image):
        # convert the boundary to a Qt polygon defining the shape in
        # relative units
        if self['Iptc4xmpExt:rbShape'] == 'rectangle':
            # polygon is two opposite corners of rectangle
            rect = QtCore.QRectF(
                self['Iptc4xmpExt:rbX'], self['Iptc4xmpExt:rbY'],
                self['Iptc4xmpExt:rbW'], self['Iptc4xmpExt:rbH'])
            polygon = QtGui.QPolygonF([rect.topLeft(), rect.bottomRight()])
        elif self['Iptc4xmpExt:rbShape'] == 'circle':
            # polygon is centre and a point on the circumference
            centre = QtCore.QPointF(
                self['Iptc4xmpExt:rbX'], self['Iptc4xmpExt:rbY'])
            edge = centre + QtCore.QPointF(self['Iptc4xmpExt:rbRx'], 0.0)
            polygon = QtGui.QPolygonF([centre, edge])
        else:
            # polygon is a list of vertices
            polygon = QtGui.QPolygonF([
                QtCore.QPointF(v['Iptc4xmpExt:rbX'], v['Iptc4xmpExt:rbY'])
                for v in self['Iptc4xmpExt:rbVertices']])
        if self['Iptc4xmpExt:rbUnit'] == 'relative':
            return polygon
        image_dims = image.metadata.dimensions
        transform = QtGui.QTransform().scale(1.0 / float(image_dims['width']),
                                             1.0 / float(image_dims['height']))
        return transform.map(polygon)

    def from_Qt(self, polygon, image):
        # convert a Qt polygon defining the shape in relative units to a
        # boundary in pixel or relative units
        boundary = dict(self)
        if boundary['Iptc4xmpExt:rbUnit'] == 'pixel':
            image_dims = image.metadata.dimensions
            transform = QtGui.QTransform().scale(float(image_dims['width']),
                                                 float(image_dims['height']))
            polygon = transform.map(polygon)
        if boundary['Iptc4xmpExt:rbShape'] == 'rectangle':
            # polygon is two opposite corners of rectangle
            rect = QtCore.QRectF(polygon.at(0), polygon.at(1))
            boundary['Iptc4xmpExt:rbX'] = rect.x()
            boundary['Iptc4xmpExt:rbY'] = rect.y()
            boundary['Iptc4xmpExt:rbW'] = rect.width()
            boundary['Iptc4xmpExt:rbH'] = rect.height()
        elif boundary['Iptc4xmpExt:rbShape'] == 'circle':
            # polygon is centre and a point on the circumference
            centre = polygon.at(0)
            radius = (polygon.at(1) - centre).manhattanLength()
            boundary['Iptc4xmpExt:rbX'] = centre.x()
            boundary['Iptc4xmpExt:rbY'] = centre.y()
            boundary['Iptc4xmpExt:rbRx'] = radius
        else:
            # polygon is a list of vertices
            boundary['Iptc4xmpExt:rbVertices'] = [
                {'Iptc4xmpExt:rbX': p.x(), 'Iptc4xmpExt:rbY': p.y()}
                for p in (polygon.at(n) for n in range(polygon.count()))]
        return RegionBoundary(boundary)


class ImageRegionItem(MD_Structure):
    extendable = True
    item_type = {
        'Iptc4xmpExt:RegionBoundary': RegionBoundary,
        'Iptc4xmpExt:rId': MD_String,
        'Iptc4xmpExt:Name': MD_LangAlt,
        'Iptc4xmpExt:rCtype': EntityConceptArray,
        'Iptc4xmpExt:rRole': EntityConceptArray,
        'Iptc4xmpExt:PersonInImage': MD_MultiString,
        'Iptc4xmpExt:OrganisationInImageName': MD_MultiString,
        'photoshop:CaptionWriter': MD_String,
        'dc:description': MD_LangAlt,
        }

    @classmethod
    def from_exiv2(cls, file_value, tag):
        if not file_value:
            return None
        if tag.startswith('Xmp'):
            return cls(file_value)
        # Convert Exif.Photo.SubjectArea to an image region. See
        # https://www.iptc.org/std/photometadata/documentation/userguide/#_mapping_exif_subjectarea_iptc_image_region
        if len(file_value) == 2:
            region = {'Iptc4xmpExt:rbShape': 'polygon',
                      'Iptc4xmpExt:rbVertices': [{
                          'Iptc4xmpExt:rbX': file_value[0],
                          'Iptc4xmpExt:rbY': file_value[1]}]}
        elif len(file_value) == 3:
            region = {'Iptc4xmpExt:rbShape': 'circle',
                      'Iptc4xmpExt:rbX': file_value[0],
                      'Iptc4xmpExt:rbY': file_value[1],
                      'Iptc4xmpExt:rbRx': file_value[2] // 2}
        elif len(file_value) == 4:
            region = {'Iptc4xmpExt:rbShape': 'rectangle',
                      'Iptc4xmpExt:rbX': file_value[0] - (file_value[2] // 2),
                      'Iptc4xmpExt:rbY': file_value[1] - (file_value[3] // 2),
                      'Iptc4xmpExt:rbW': file_value[2],
                      'Iptc4xmpExt:rbH': file_value[3]}
        else:
            return None
        region['Iptc4xmpExt:rbUnit'] = 'pixel'
        return cls({
            'Iptc4xmpExt:RegionBoundary': region,
            'Iptc4xmpExt:rRole': [{
                'Iptc4xmpExt:Name': {'en-GB': 'main subject area'},
                'xmp:Identifier': [
                    'http://cv.iptc.org/newscodes/imageregionrole/mainSubjectArea'],
                }],
            })

    def has_uid(self, key, uid):
        if key not in self:
            return False
        for item in self[key]:
            if 'xmp:Identifier' in item and uid in item['xmp:Identifier']:
                return True
        return False

    def is_main_subject_area(self):
        return self.has_uid(
            'Iptc4xmpExt:rRole',
            'http://cv.iptc.org/newscodes/imageregionrole/mainSubjectArea')

    def to_Qt(self, image):
        return self['Iptc4xmpExt:RegionBoundary'].to_Qt(image)

    def from_Qt(self, polygon, image):
        return self['Iptc4xmpExt:RegionBoundary'].from_Qt(polygon, image)

    def convert_unit(self, unit, image):
        boundary = self['Iptc4xmpExt:RegionBoundary']
        if boundary['Iptc4xmpExt:rbUnit'] == unit:
            return self
        polygon = boundary.to_Qt(image)
        boundary['Iptc4xmpExt:rbUnit'] = unit
        boundary = boundary.from_Qt(polygon, image)
        result = ImageRegionItem(self)
        result['Iptc4xmpExt:RegionBoundary'] = boundary
        return result


class MD_ImageRegion(MD_StructArray):
    item_type = ImageRegionItem

    def new_region(self, region, idx=None):
        if idx is None:
            idx = len(self)
        result = list(self)
        if region:
            if idx < len(self):
                result[idx] = region
            else:
                result.append(region)
        elif idx < len(self):
            result.pop(idx)
        result = MD_ImageRegion(result)
        return result

    def index(self, other):
        if other.is_main_subject_area():
            # only one main subject area region allowed
            for n, value in enumerate(self):
                if value.is_main_subject_area():
                    return True
            return len(self)
        for n, value in enumerate(self):
            for key in ('Iptc4xmpExt:RegionBoundary', 'Iptc4xmpExt:rId'):
                if value[key] and value[key] == other[key]:
                    return n
        return len(self)

    @staticmethod
    def boundary_from_note(note, dims, image):
        if not ('x' in note and 'y' in note and
                'w' in note and 'h' in note):
            return None
        transform = (image.metadata.orientation
                     and image.metadata.orientation.get_transform())
        w, h = dims
        if transform and transform.isRotating():
            w, h = h, w
        scale = QtGui.QTransform().scale(1.0 / float(w), 1.0 / float(h))
        rect = QtCore.QRectF(float(note['x']), float(note['y']),
                             float(note['w']), float(note['h']))
        rect = scale.mapRect(rect)
        if transform:
            rect = transform.inverted()[0].mapRect(rect)
        boundary = {'Iptc4xmpExt:rbShape': 'rectangle',
                    'Iptc4xmpExt:rbUnit': 'relative'}
        (boundary['Iptc4xmpExt:rbX'],
         boundary['Iptc4xmpExt:rbY'],
         boundary['Iptc4xmpExt:rbW'],
         boundary['Iptc4xmpExt:rbH']) = rect.getRect()
        return boundary

    def from_notes(self, notes, image, target_size):
        # convert current regions to note boundaries for comparison
        boundaries = {}
        for region, note in self.to_note_boundary(image, target_size):
            key = ','.join(note[k] for k in ('x', 'y', 'w', 'h'))
            boundaries[key] = region['Iptc4xmpExt:RegionBoundary']
        dims = image.metadata.dimensions.scaled_to(target_size)
        result = []
        for note in notes:
            # use existing boundary if it matches (to cope with low
            # resolution of Flickr / Ipernity)
            key = ','.join(note[k] for k in ('x', 'y', 'w', 'h'))
            if key in boundaries:
                boundary = dict(boundaries[key])
            else:
                boundary = self.boundary_from_note(note, dims, image)
            if not boundary:
                continue
            region = {'Iptc4xmpExt:RegionBoundary': boundary}
            region['Iptc4xmpExt:rRole'] = [{
                'xmp:Identifier': [
                    'http://cv.iptc.org/newscodes/imageregionrole/subjectArea'],
                'Iptc4xmpExt:Name': {'en-GB': 'subject area'}}]
            if note['is_person']:
                region['Iptc4xmpExt:PersonInImage'] = [note['content']]
                region['Iptc4xmpExt:rCtype'] = [{
                    'Iptc4xmpExt:Name': {'en-GB': 'human'},
                    'xmp:Identifier': [
                        'http://cv.iptc.org/newscodes/imageregiontype/human']}]
            else:
                region['dc:description'] = {'x-default': note['content']}
            if note['authorrealname']:
                region['photoshop:CaptionWriter'] = note['authorrealname']
            result.append(region)
        return result

    def to_note_boundary(self, image, target_size):
        w, h = image.metadata.dimensions.scaled_to(target_size)
        transform = (image.metadata.orientation
                     and image.metadata.orientation.get_transform())
        if transform and transform.isRotating():
            w, h = h, w
        scale = QtGui.QTransform().scale(w, h)
        for region in self:
            boundary = region['Iptc4xmpExt:RegionBoundary']
            if boundary['Iptc4xmpExt:rbShape'] != 'rectangle':
                continue
            points = region.to_Qt(image)
            rect = QtCore.QRectF(points.at(0), points.at(1))
            if transform:
                rect = transform.mapRect(rect).normalized()
            rect = scale.mapRect(rect)
            note = dict(zip(('x', 'y', 'w', 'h'),
                            [str(int(v + 0.5)) for v in rect.getRect()]))
            yield region, note

    def to_notes(self, image, target_size):
        result = []
        for region, note in self.to_note_boundary(image, target_size):
            note['content'] = ''
            note['is_person'] = False
            if region.has_uid(
                    'Iptc4xmpExt:rCtype',
                    'http://cv.iptc.org/newscodes/imageregiontype/human'):
                if 'Iptc4xmpExt:PersonInImage' in region:
                    note['content'] = ', '.join(
                        region['Iptc4xmpExt:PersonInImage'])
                note['is_person'] = True
            elif not any(region.has_uid('Iptc4xmpExt:rRole', x) for x in (
                'http://cv.iptc.org/newscodes/imageregionrole/subjectArea',
                'http://cv.iptc.org/newscodes/imageregionrole/mainSubjectArea',
                'http://cv.iptc.org/newscodes/imageregionrole/areaOfInterest')):
                continue
            if 'dc:description' in region and not note['content']:
                note['content'] = MD_LangAlt(
                    region['dc:description']).best_match()
            if 'Iptc4xmpExt:Name' in region and not note['content']:
                note['content'] = MD_LangAlt(
                    region['Iptc4xmpExt:Name']).best_match()
            if not note['content']:
                continue
            result.append(note)
        return result

    def get_focus(self, image):
        image_dims = image.metadata.dimensions
        portrait_format = image_dims['height'] > image_dims['width']
        transform = (image.metadata.orientation
                     and image.metadata.orientation.get_transform())
        if transform and transform.isRotating():
            portrait_format = not portrait_format
        if portrait_format:
            uris = (
                'http://cv.iptc.org/newscodes/imageregionrole/landscapeCropping',
                'http://cv.iptc.org/newscodes/imageregionrole/squareCropping',
                'http://cv.iptc.org/newscodes/imageregionrole/recomCropping',
                'http://cv.iptc.org/newscodes/imageregionrole/cropping',
                'http://cv.iptc.org/newscodes/imageregionrole/portraitCropping',
                )
        else:
            uris = (
                'http://cv.iptc.org/newscodes/imageregionrole/squareCropping',
                'http://cv.iptc.org/newscodes/imageregionrole/portraitCropping',
                'http://cv.iptc.org/newscodes/imageregionrole/landscapeCropping',
                'http://cv.iptc.org/newscodes/imageregionrole/recomCropping',
                'http://cv.iptc.org/newscodes/imageregionrole/cropping',
                )
        for uri in uris:
            for region in self:
                if not region.has_uid('Iptc4xmpExt:rRole', uri):
                    continue
                points = region.to_Qt(image)
                boundary = region['Iptc4xmpExt:RegionBoundary']
                if boundary['Iptc4xmpExt:rbShape'] == 'rectangle':
                    centre = (points.at(0) + points.at(1)) / 2.0
                elif boundary['Iptc4xmpExt:rbShape'] == 'circle':
                    centre = points.at(0)
                else:
                    centre = points.boundingRect().center()
                if transform:
                    centre = transform.map(centre)
                return (centre.x() * 2.0) - 1.0, 1.0 - (centre.y() * 2.0)
        return None



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import json
import subprocess
import sys


def startupinfo():
    if sys.platform.startswith('win'):
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        return startupinfo
    return None

try:
    ffmpeg_version = subprocess.check_output(
        ['ffprobe', '-hide_banner', '-loglevel', 'warning', '-version'],
        startupinfo=startupinfo())
    ffmpeg_version = subprocess.check_output(
        ['ffmpeg', '-hide_banner', '-loglevel', 'warning', '-version'],
        startupinfo=startupinfo())
    ffmpeg_version = ffmpeg_version.decode('utf-8')
    ffmpeg_version = ffmpeg_version.splitlines()[0]
except OSError as ex:
    print('ffmpeg or ffprobe not found')
    ffmpeg_version = None


class FFmpeg(object):
    @staticmethod
    def ffprobe(path, options=['-show_format', '-show_streams']):
        if not ffmpeg_version:
            return {}
        cmd = ['ffprobe', '-hide_banner', '-loglevel', 'warning']
        cmd += options
        cmd += ['-select_streams', 'v', '-print_format', 'json', path]
        p = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            startupinfo=startupinfo())
        output, error = p.communicate()
        if p.returncode:
            error = error.decode('utf-8')
            error = error.splitlines()[0]
            raise RuntimeError('ffprobe: {}'.format(error))
        output = output.decode('utf-8')
        return json.loads(output)

    @staticmethod
    def make_thumbnail(path, w, h, skip, quality):
        if not ffmpeg_version:
            return None
        cmd = ['ffmpeg', '-hide_banner', '-loglevel', 'warning']
        if skip > 0:
            cmd += ['-ss', str(skip)]
        cmd += ['-noautorotate', '-i', path, '-an', '-vframes', '1']
        cmd += ['-vf', ('scale={w}:{h}:force_original_aspect_ratio=decrease,'
                        'pad={w}:{h}:(ow-iw)/2:(oh-ih)/2').format(w=w, h=h)]
        cmd += ['-sws_flags', 'sinc', '-f', 'image2pipe',
                '-vcodec', 'mjpeg', '-q:v', str(quality), 'pipe:1']
        p = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            startupinfo=startupinfo())
        output, error = p.communicate()
        if p.returncode:
            error = error.decode('utf-8')
            error = error.splitlines()[0]
            raise RuntimeError('ffmpeg: {}'.format(error))
        return output



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2019-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from collections import defaultdict
import locale
import logging
import os

import requests

from photini.configstore import key_store
from photini.metadata import ImageMetadata
from photini.photinimap import GeocoderBase
from photini.pyqt import *
from photini.types import MD_Location
from photini.widgets import (
    CompactButton, DoubleSpinBox, LatLongDisplay, LangAltWidget, SingleLineEdit)

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class OpenCage(GeocoderBase):
    api_key = key_store.get('opencage', 'api_key')

    def query(self, params):
        params['key'] = self.api_key
        params['no_annotations'] = '1'
        with Busy():
            self.rate_limit()
            try:
                rsp = requests.get(
                    'https://api.opencagedata.com/geocode/v1/json',
                    params=params, timeout=5)
            except Exception as ex:
                logger.error(str(ex))
                return []
        if rsp.status_code >= 400:
            logger.error('Search error %d', rsp.status_code)
            return []
        rsp = rsp.json()
        status = rsp['status']
        if status['code'] != 200:
            logger.error(
                'Search error %d: %s', status['code'], status['message'])
            return []
        if rsp['total_results'] < 1:
            logger.error('No results found')
            return []
        rate = rsp['rate']
        self.block_timer.setInterval(
            5000 * rate['limit'] // max(rate['remaining'], 1))
        return rsp['results']

    # Map OpenCage address components to IPTC address heirarchy. There
    # are many possible components (user generated data) so any
    # unrecognised ones are put in 'Iptc4xmpExt:Sublocation'. See
    # https://github.com/OpenCageData/address-formatting/blob/master/conf/components.yaml
    address_map = {
        'Iptc4xmpExt:WorldRegion': ('continent',),
        'Iptc4xmpExt:CountryCode': (
            'ISO_3166-1_alpha-3', 'ISO_3166-1_alpha-2', 'country_code'),
        'Iptc4xmpExt:CountryName': ('country', 'country_name'),
        'Iptc4xmpExt:ProvinceState': (
            'county', 'county_code', 'local_administrative_area',
            'state_district', 'state', 'state_code', 'province',
            'region', 'island'),
        'Iptc4xmpExt:City': (
            'neighbourhood', 'city_block', 'quarter', 'suburb', 'district',
            'borough', 'city_district', 'commercial', 'industrial', 'houses',
            'subdivision', 'village', 'town', 'municipality', 'city',
            'partial_postcode', 'postcode'),
        'Iptc4xmpExt:Sublocation': (
            'house_number', 'street_number', 'house', 'public_building',
            'building', 'residential', 'water', 'road', 'pedestrian', 'path',
            'street_name', 'street', 'cycleway', 'footway', 'place', 'square',
            'locality', 'hamlet', 'croft'),
        'ignore': (
            'ISO_3166-2', 'political_union', 'road_reference',
            'road_reference_intl', 'road_type', '_category', '_type'),
        }

    def get_address(self, coords):
        params = {'q': '{:.5f},{:.5f}'.format(*coords)}
        lang, encoding = locale.getdefaultlocale()
        if lang:
            params['language'] = lang
        results = self.cached_query(params)
        if not results:
            return None
        address = dict(results[0]['components'])
        formatted = results[0]['formatted']
        for key in address.keys():
            if isinstance(address[key], list):
                try:
                    address[key] = '; '.join(address[key])
                except Exception:
                    pass
            if not isinstance(address[key], str):
                try:
                    address[key] = str(address[key])
                except Exception:
                    del address[key]
        # remove some known equivalent data
        for key_1, key_2 in (('county_code', 'county'),
                             ('state_code', 'state'),
                             ('partial_postcode', 'postcode')):
            if key_1 in address and key_2 in address:
                del address[key_1]
        # remove duplicate data
        for key_list in self.address_map.values():
            for n, key_1 in enumerate(key_list):
                if key_1 not in address:
                    continue
                for key_2 in key_list[n+1:]:
                    if key_2 not in address:
                        continue
                    if address[key_1] == address[key_2]:
                        del address[key_1]
                        break
        # attempt to format postcode correctly
        for key in 'city', 'town', 'village', 'suburb', 'state', 'country':
            if 'postcode' in address and key in address:
                for fmt in '{0} {1}', '{0}, {1}', '{1} {0}', '{1}, {0}':
                    guess = fmt.format(address['postcode'], address[key])
                    if guess in formatted:
                        address[key] = guess
                        del address['postcode']
                        break
        gps = results[0]['geometry']
        return MD_Location.from_address(gps, address, self.address_map)

    def search_terms(self):
        text = translate('AddressTab', 'Address lookup powered by OpenCage')
        tou_opencage = CompactButton(text)
        tou_opencage.clicked.connect(self.load_tou_opencage)
        tou_osm = CompactButton(
            translate('AddressTab', 'Geodata © OpenStreetMap contributors'))
        tou_osm.clicked.connect(self.load_tou_osm)
        return [tou_opencage, tou_osm]

    @QtSlot()
    @catch_all
    def load_tou_opencage(self):
        QtGui.QDesktopServices.openUrl(
            QtCore.QUrl('https://geocoder.opencagedata.com/'))

    @QtSlot()
    @catch_all
    def load_tou_osm(self):
        QtGui.QDesktopServices.openUrl(
            QtCore.QUrl('http://www.openstreetmap.org/copyright'))


class LocationInfo(QtWidgets.QScrollArea):
    new_value = QtSignal(object, dict)

    def __init__(self, *args, **kw):
        super(LocationInfo, self).__init__(*args, **kw)
        self.setFrameStyle(QtWidgets.QFrame.Shape.NoFrame)
        self.setWidgetResizable(True)
        form = QtWidgets.QWidget()
        layout = QtWidgets.QGridLayout()
        form.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        self.members = {}
        self.members['LocationName'] = LangAltWidget(
            'Iptc4xmpExt:LocationName', multi_line=False)
        self.members['LocationName'].setToolTip('<p>{}</p>'.format(
            translate('AddressTab', 'Enter a full name of the location.')))
        self.members['LocationName'].new_value.connect(self.editing_finished)
        for (key, tool_tip) in (
                ('Sublocation', translate(
                    'AddressTab', 'Enter the name of the sublocation.')),
                ('City', translate(
                    'AddressTab', 'Enter the name of the city.')),
                ('ProvinceState', translate(
                    'AddressTab', 'Enter the name of the province or state.')),
                ('CountryName', translate(
                    'AddressTab', 'Enter the name of the country.')),
                ('CountryCode', translate(
                    'AddressTab', 'Enter the 2 or 3 letter ISO 3166 country'
                    ' code of the country.')),
                ('WorldRegion', translate(
                    'AddressTab', 'Enter the name of the world region.')),
                ('LocationId', translate(
                    'AddressTab', 'Enter globally unique identifier(s) of the'
                    ' location. Separate them with ";" characters.'))):
            self.members[key] = SingleLineEdit(
                'Iptc4xmpExt:' + key, length_check=ImageMetadata.max_bytes(key))
            self.members[key].setToolTip('<p>{}</p>'.format(tool_tip))
            self.members[key].new_value.connect(self.editing_finished)
        self.members['latlon'] = LatLongDisplay()
        self.members['latlon'].new_value.connect(self.editing_finished)
        self.members['alt'] = DoubleSpinBox('exif:GPSAltitude')
        self.members['alt'].set_suffix(
            translate('AddressTab', ' m', 'metres altitude'))
        self.members['alt'].setToolTip('<p>{}</p>'.format(
            translate('AddressTab', 'Altitude of the location in metres.')))
        self.members['alt'].new_value.connect(self.editing_finished)
        self.members['CountryCode'].setMaximumWidth(
            width_for_text(self.members['CountryCode'], 'W' * 4))
        for j, text in enumerate((translate('AddressTab', 'Name'),
                                  translate('AddressTab', 'Street'),
                                  translate('AddressTab', 'City'),
                                  translate('AddressTab', 'Province'),
                                  translate('AddressTab', 'Country'),
                                  translate('AddressTab', 'Region'),
                                  translate('AddressTab', 'Location ID'))):
            label = QtWidgets.QLabel(text)
            label.setAlignment(Qt.AlignmentFlag.AlignRight)
            layout.addWidget(label, j, 0)
        layout.addWidget(self.members['LocationName'], 0, 1, 1, 5)
        layout.addWidget(self.members['Sublocation'], 1, 1, 1, 5)
        layout.addWidget(self.members['City'], 2, 1, 1, 5)
        layout.addWidget(self.members['ProvinceState'], 3, 1, 1, 5)
        layout.addWidget(self.members['CountryName'], 4, 1, 1, 4)
        layout.addWidget(self.members['CountryCode'], 4, 5)
        layout.addWidget(self.members['WorldRegion'], 5, 1, 1, 5)
        layout.addWidget(self.members['LocationId'], 6, 1, 1, 5)
        layout.addWidget(self.members['latlon'].label, 7, 0)
        layout.addWidget(self.members['latlon'], 7, 1)
        label = QtWidgets.QLabel(translate('AddressTab', 'Altitude'))
        label.setAlignment(Qt.AlignmentFlag.AlignRight)
        layout.addWidget(label, 7, 2)
        self.members['alt'].setFixedWidth(self.members['latlon'].width())
        layout.addWidget(self.members['alt'], 7, 3)
        layout.setColumnStretch(4, 1)
        layout.setRowStretch(8, 1)
        self.setWidget(form)

    def get_value(self):
        new_value = {}
        for key in self.members:
            new_value.update(self.members[key].get_value_dict())
        return new_value

    @QtSlot(dict)
    @catch_all
    def editing_finished(self, value):
        self.new_value.emit(self, value)


class QTabBar(QtWidgets.QTabBar):
    context_menu = QtSignal(QtGui.QContextMenuEvent)

    @catch_all
    def contextMenuEvent(self, event):
        self.context_menu.emit(event)


class TabWidget(QtWidgets.QWidget):
    @staticmethod
    def tab_name():
        return translate('AddressTab', '&Address')

    def __init__(self, parent=None):
        super(TabWidget, self).__init__(parent)
        self.app = QtWidgets.QApplication.instance()
        self.geocoder = OpenCage(parent=self)
        self.setLayout(QtWidgets.QHBoxLayout())
        ## left side
        left_side = QtWidgets.QGridLayout()
        # latitude & longitude
        self.coords = LatLongDisplay()
        self.coords.setReadOnly(True)
        left_side.addWidget(self.coords.label, 0, 0)
        left_side.addWidget(self.coords, 0, 1)
        # convert lat/lng to location info
        self.auto_location = QtWidgets.QPushButton(
            translate('AddressTab', 'Get address from lat, long'))
        self.auto_location.setEnabled(False)
        self.auto_location.clicked.connect(self.get_address)
        left_side.addWidget(self.auto_location, 1, 0, 1, 2)
        # terms and conditions
        terms = self.geocoder.search_terms()
        left_side.addWidget(terms[0], 3, 0, 1, 2)
        left_side.addWidget(terms[1], 4, 0, 1, 2)
        left_side.setColumnStretch(1, 1)
        left_side.setRowStretch(2, 1)
        self.layout().addLayout(left_side)
        ## right side
        # location info
        self.location_widgets = []
        self.location_info = QtWidgets.QTabWidget()
        tab_bar = QTabBar()
        self.location_info.setTabBar(tab_bar)
        tab_bar.context_menu.connect(self.location_tab_context_menu)
        tab_bar.tabMoved.connect(self.location_tab_moved)
        self.location_info.setElideMode(Qt.TextElideMode.ElideLeft)
        self.location_info.setMovable(True)
        self.location_info.setEnabled(False)
        self.layout().addWidget(self.location_info, stretch=1)

    def refresh(self):
        self.new_selection(self.app.image_list.get_selected_images())

    def do_not_close(self):
        return False

    @QtSlot(QtGui.QContextMenuEvent)
    @catch_all
    def location_tab_context_menu(self, event):
        idx = self.location_info.tabBar().tabAt(event.pos())
        self.location_info.setCurrentIndex(idx)
        menu = QtWidgets.QMenu(self)
        menu.addAction(translate(
            'AddressTab', 'Duplicate location'), self.duplicate_location)
        menu.addAction(translate(
            'AddressTab', 'Delete location'), self.delete_location)
        action = execute(menu, event.globalPos())

    @QtSlot()
    @catch_all
    def duplicate_location(self):
        idx = self.location_info.currentIndex()
        for image in self.app.image_list.get_selected_images():
            # duplicate data
            location = self._get_location(image, idx)
            # shuffle data up
            location_list = list(image.metadata.location_shown)
            location_list.insert(idx, location)
            image.metadata.location_shown = location_list
        # display data
        self.display_location()

    @QtSlot()
    @catch_all
    def delete_location(self):
        idx = self.location_info.currentIndex()
        for image in self.app.image_list.get_selected_images():
            # shuffle data down
            location_list = list(image.metadata.location_shown)
            if idx == 0:
                if location_list:
                    location = [location_list.pop(0)]
                else:
                    location = []
                image.metadata.location_taken = location
            elif idx <= len(location_list):
                del location_list[max(idx - 1, 0)]
            image.metadata.location_shown = location_list
        # display data
        self.display_location()

    @QtSlot(int, int)
    @catch_all
    def location_tab_moved(self, idx_a, idx_b):
        self.pending_move = idx_a, idx_b
        # do actual swap when idle to avoid seg fault
        QtCore.QTimer.singleShot(0, self._location_tab_moved)

    @QtSlot()
    @catch_all
    def _location_tab_moved(self):
        idx_a, idx_b = self.pending_move
        # swap data
        for image in self.app.image_list.get_selected_images():
            temp_a = self._get_location(image, idx_a)
            temp_b = self._get_location(image, idx_b)
            self._set_location(image, idx_a, temp_b)
            self._set_location(image, idx_b, temp_a)
        # adjust tab names
        for idx in range(min(idx_a, idx_b), max(idx_a, idx_b) + 1):
            self.set_tab_text(idx)
        # display data
        self.display_location()

    def _get_location(self, image, idx):
        if idx == 0:
            if image.metadata.location_taken:
                return image.metadata.location_taken[0]
            return {}
        elif idx <= len(image.metadata.location_shown):
            return image.metadata.location_shown[idx - 1]
        return {}

    def _set_location(self, image, idx, location):
        if idx == 0:
            image.metadata.location_taken = [location]
        else:
            location_list = list(image.metadata.location_shown)
            while len(location_list) < idx:
                location_list.append(None)
            location_list[idx - 1] = location
            image.metadata.location_shown = location_list

    @QtSlot(object, dict)
    @catch_all
    def new_location(self, widget, new_value, images=[]):
        images = images or self.app.image_list.get_selected_images()
        idx = self.location_info.indexOf(widget)
        for image in images:
            temp = dict(self._get_location(image, idx))
            temp.update(new_value)
            self._set_location(image, idx, temp)
        # new_location can be called when changing tab, so don't delete
        # tabs until later
        QtCore.QTimer.singleShot(0, self.display_location)

    def set_tab_text(self, idx):
        if idx == 0:
            text = translate('AddressTab', 'camera')
            tip = translate('AddressTab', 'Enter the details about a location'
                            ' where this image was created.')
        else:
            text = translate('AddressTab', 'subject {idx}').format(idx=idx)
            tip = translate('AddressTab', 'Enter the details about a location'
                            ' which is shown in this image.')
        self.location_info.setTabText(idx, text)
        self.location_info.setTabToolTip(idx, '<p>' + tip + '</p>')

    @QtSlot()
    @catch_all
    def display_location(self):
        images = self.app.image_list.get_selected_images()
        # get required number of tabs
        count = 0
        for image in images:
            if image.metadata.location_shown:
                count = max(count, len(image.metadata.location_shown))
        count += 2
        # add or remove tabs
        if self.location_info.currentIndex() >= count:
            self.location_info.setCurrentIndex(count - 1)
        idx = self.location_info.count()
        while idx < count:
            if not self.location_widgets:
                widget = LocationInfo()
                widget.new_value.connect(self.new_location)
                self.location_widgets.append(widget)
            self.location_info.addTab(self.location_widgets.pop(), '')
            self.set_tab_text(idx)
            idx += 1
        while idx > count:
            idx -= 1
            self.location_widgets.append(self.location_info.widget(idx))
            self.location_info.removeTab(idx)
        # display data
        for idx in range(count):
            widget = self.location_info.widget(idx)
            values = [self._get_location(image, idx) for image in images]
            for key in widget.members:
                widget.members[key].set_value_list(values)

    def new_selection(self, selection):
        self.location_info.setEnabled(bool(selection))
        values = []
        for image in selection:
            values.append(image.metadata.gps_info)
        self.coords.set_value_list(values)
        self.auto_location.setEnabled(bool(self.coords.get_value()))
        self.display_location()

    @QtSlot()
    @catch_all
    def get_address(self):
        images = self.app.image_list.get_selected_images()
        location = self.geocoder.get_address(self.coords.get_value())
        if location:
            self.new_location(
                self.location_info.currentWidget(), location, images)



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2023  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import itertools
import logging
import math
import os
import re

import requests
from requests_oauthlib import OAuth2Session
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from photini.configstore import BaseConfigStore, key_store
from photini.pyqt import *
from photini.uploader import (
    PhotiniUploader, UploadAborted, UploaderSession, UploaderUser)
from photini.widgets import (
    DropDownSelector, Label, MultiLineEdit, SingleLineEdit)

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate

# Pixelfed API: https://docs.pixelfed.org/technical-documentation/api/
# Endpoint list: https://github.com/pixelfed/pixelfed/blob/dev/routes/api.php


class PixelfedSession(UploaderSession):
    def open_connection(self):
        if self.api:
            return
        auto_refresh_kwargs = {
            'client_id': self.client_data['client_id'],
            'client_secret': self.client_data['client_secret'],
            }
        self.api = OAuth2Session(
            client_id=self.client_data['client_id'],
            token=self.user_data['token'],
            auto_refresh_url=self.client_data['api_base_url'] + '/oauth/token',
            auto_refresh_kwargs=auto_refresh_kwargs,
            token_updater=self.save_token)
        self.api.headers.update(self.headers)

    def save_token(self, token):
        self.user_data['token'] = token
        self.new_token.emit(token)

    @staticmethod
    def check_response(rsp, decode=True):
        UploaderSession.check_interrupt()
        try:
            rsp.raise_for_status()
            # some wrong api calls return the instance home page
            if rsp.headers['Content-Type'].startswith('text/html'):
                logger.error('Response is HTML')
                return None
            if decode:
                rsp = rsp.json()
        except UploadAborted:
            raise
        except Exception as ex:
            logger.error(str(ex))
            return None
        return rsp

    def api_call(self, endpoint, method='GET', **params):
        self.open_connection()
        url = self.client_data['api_base_url'] + endpoint
        rsp = self.check_response(self.api.request(method, url, **params))
        if rsp is None:
            self.close_connection()
        elif 'error' in rsp:
            logger.error('%s: %s', endpoint, rsp['error'])
            return None
        return rsp

    def upload_files(self, upload_list):
        media_ids = []
        licence = None
        upload_count = 0
        for image, convert, params in upload_list:
            upload_count += 1
            self.upload_progress.emit({
                'label': '{} ({}/{})'.format(os.path.basename(image.path),
                                             upload_count, len(upload_list)),
                'busy': True})
            default_licence = params['default_license']
            licence = licence or default_licence
            # set licence
            if params['license'] != licence:
                data = {'license': params['license']}
                rsp = self.api_call(
                    '/api/v1/accounts/update_credentials', 'PATCH', data=data)
                if rsp:
                    licence = params['license']
            # upload media
            fields = dict(params['media'])
            with self.open_file(image, convert) as (image_type, fileobj):
                fields['file'] = (params['file_name'], fileobj, image_type)
                data = MultipartEncoderMonitor(
                    MultipartEncoder(fields=fields), self.progress)
                self.upload_progress.emit({'busy': False})
                endpoint = '/api/v1/media'
                if (self.client_data['version']['mastodon'] >= (3, 1, 3) or
                    (self.client_data['version']['pixelfed'] and
                     self.client_data['version']['pixelfed'] >= (0, 11, 3))):
                    endpoint = '/api/v2/media'
                rsp = self.api_call(
                    endpoint, method='POST', data=data,
                    headers={'Content-Type': data.content_type})
            self.upload_progress.emit({'busy': True})
            if rsp is None:
                self.upload_progress.emit({
                    'error': (image, 'Image upload failed')})
                return
            media_ids.append((image, rsp['id']))
        # reset licence
        if licence != default_licence:
            data = {'license': default_licence}
            rsp = self.api_call(
                '/api/v1/accounts/update_credentials', 'PATCH', data=data)
        # upload status
        data = dict(params['status'])
        data['media_ids[]'] = [x[1] for x in media_ids]
        rsp = self.api_call(
            '/api/v1/statuses', method='POST', data=data)
        if rsp is None:
            self.upload_progress.emit({'error': (image, 'Post status failed')})
            return
        # store photo ids in image keywords, in main thread
        for image, media_id in media_ids:
            self.upload_progress.emit({
                'keyword': (image, '{}:id={}'.format(
                    self.client_data['name'], media_id))})


class ChooseInstance(QtWidgets.QDialog):
    def __init__(self, default=None, instances=[], **kw):
        super(ChooseInstance, self).__init__(**kw)
        self.setWindowTitle(translate(
            'PixelfedTab', 'Photini: choose instance'))
        self.setLayout(QtWidgets.QVBoxLayout())
        # text
        self.layout().addWidget(QtWidgets.QLabel(
            '<h3>{}</h3>'.format(translate(
            'PixelfedTab', 'Choose an instance'))))
        self.layout().addWidget(QtWidgets.QLabel(translate(
            'PixelfedTab', 'Which Pixelfed instance hosts your account?')))
        # list of instances
        scroll_area = QtWidgets.QScrollArea()
        scroll_area.setWidgetResizable(True)
        self.layout().addWidget(scroll_area)
        panel = QtWidgets.QWidget()
        panel.setLayout(QtWidgets.QGridLayout())
        self.buttons = []
        row = 0
        for instance in instances:
            button = QtWidgets.QRadioButton(instance)
            button.setChecked(instance == default)
            self.buttons.append(button)
            panel.layout().addWidget(button, row, 0, 1, 2)
            row += 1
        # any other instance
        button = QtWidgets.QRadioButton(translate('PixelfedTab', 'Other'))
        self.buttons.append(button)
        button.setEnabled(False)
        panel.layout().addWidget(button, row, 0)
        self.other_text = QtWidgets.QLineEdit()
        self.other_text.textChanged.connect(self.text_changed)
        panel.layout().addWidget(self.other_text, row, 1)
        # add panel to scroll area now its size is known
        scroll_area.setWidget(panel)
        # ok & cancel buttons
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        self.ok_button = button_box.addButton(
            QtWidgets.QDialogButtonBox.StandardButton.Ok)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        self.layout().addWidget(button_box)
        for button in self.buttons:
            button.toggled.connect(self.button_toggled)
        self.button_toggled()

    @QtSlot(str)
    @catch_all
    def text_changed(self, value):
        self.buttons[-1].setChecked(bool(value))
        self.buttons[-1].setEnabled(bool(value))

    @QtSlot(bool)
    @catch_all
    def button_toggled(self, value=None):
        for button in self.buttons:
            if button.isChecked():
                self.ok_button.setEnabled(True)
                return
        self.ok_button.setEnabled(False)

    def execute(self):
        if execute(self) != QtWidgets.QDialog.DialogCode.Accepted:
            return None
        for button in self.buttons:
            if button.isChecked():
                if button == self.buttons[-1]:
                    return self.other_text.text().strip().strip('/')
                return button.text()
        return None


class PixelfedUser(UploaderUser):
    new_instance_config = QtSignal(dict)
    logger = logger
    config_section = 'pixelfed'
    scopes = ['read', 'write']

    def on_connect(self, widgets):
        with self.session(parent=self) as session:
            yield 'connected', session.api.authorized
            # get user info
            account = session.api_call('/api/v1/accounts/verify_credentials')
            if not account:
                self.log_out()
                yield 'connected', False
            self.user_data['id'] = account['id']
            self.user_data['lang'] = account['source']['language']
            # get icon
            icon_url = account['avatar_static']
            rsp = session.check_response(
                session.api.get(icon_url), decode=False)
            yield 'user', (account['display_name'], rsp and rsp.content)
            # get instance info
            self.client_data['version'] = {'mastodon': None, 'pixelfed': None}
            self.instance_config = session.api_call('/api/v1/instance')
            if not self.instance_config:
                yield 'connected', False
            self.new_instance_config.emit(self.instance_config)
            widgets['status'].highlighter.length_check = self.instance_config[
                'configuration']['statuses']['max_characters']
            version = self.instance_config['version']
            logger.info('server version "%s"', version)
            match = re.match(r'(\d+)\.(\d+)\.(\d+)', version)
            if match:
                self.client_data['version']['mastodon'] = tuple(
                    int(x) for x in match.groups())
            match = re.search(r'Pixelfed\s+(\d+)\.(\d+)\.(\d+)', version)
            if match:
                self.client_data['version']['pixelfed'] = tuple(
                    int(x) for x in match.groups())
            self.unavailable['albums'] = not (
                self.client_data['version']['pixelfed']
                and self.client_data['version']['pixelfed'] >= (0, 11, 4))
            self.unavailable['comments_disabled'] = self.unavailable['albums']
            self.unavailable['license'] = not (
                self.client_data['version']['pixelfed']
                and self.client_data['version']['pixelfed'] >= (0, 11, 1))
            self.unavailable['new_album'] = not (
                self.client_data['version']['pixelfed']
                and self.client_data['version']['pixelfed'] >= (99, 0, 0))
            self.unavailable['sync_remote'] = not (
                self.client_data['version']['pixelfed']
                and self.client_data['version']['pixelfed'] >= (0, 10, 6))
            media = self.instance_config['configuration']['media_attachments']
            self.max_size = {
                'image': {'bytes': media['image_size_limit'],
                          'pixels': media['image_matrix_limit']},
                'video': {'bytes': media['video_size_limit'],
                          'pixels': media['video_matrix_limit']},
                }
            yield None, None
            # get user preferences
            prefs = session.api_call('/api/v1/preferences')
            if prefs is None:
                yield 'connected', False
            if prefs['posting:default:language']:
                self.user_data['lang'] = prefs['posting:default:language']
            widgets['sensitive'].setChecked(
                prefs['posting:default:sensitive'])
            widgets['visibility'].set_value(
                prefs['posting:default:visibility'])
            if self.client_data['version']['pixelfed']:
                self.compose_settings = session.api_call(
                    '/api/v1.1/compose/settings')
                if self.compose_settings is None:
                    yield 'connected', False
                if 'default_license' in self.compose_settings:
                    widgets['license'].set_value(
                        int(self.compose_settings['default_license']))
                if 'default_scope' in self.compose_settings:
                    widgets['visibility'].set_value(
                        self.compose_settings['default_scope'])
            else:
                self.compose_settings = {
                    'max_altext_length': 5000,
                    'default_license': widgets['license'].get_value(),
                    'media_descriptions': True,
                    }
            yield None, None
            # get collections
            if self.unavailable['albums']:
                return
            for collection in session.api_call(
                    '/api/v1.1/collections/accounts/{}'.format(
                        self.user_data['id'])) or []:
                if collection:
                    yield 'album', {
                        'title': collection['title'] or 'Untitled',
                        'description': collection['description'],
                        'id': collection['id'],
                        'writeable': collection['post_count'] < 18,
                        }

    def load_user_data(self):
        self.config_store = QtWidgets.QApplication.instance().config_store
        # get list of known instances
        self.instances = []
        self.instance_data = {}
        # registered instances
        for section in key_store.config.sections():
            name, sep, instance = section.partition(' ')
            if name != 'pixelfed':
                continue
            instance_data = {'name': instance,
                             'api_base_url': 'https://' + instance}
            for option in key_store.config.options(section):
                instance_data[option] = key_store.get(section, option)
            self.instances.append(instance)
            self.instance_data[instance] = instance_data
        # locally registered instances
        self.local_config = BaseConfigStore('pixelfed')
        for instance in self.local_config.config.sections():
            if instance in self.instances:
                continue
            instance_data = {'name': instance,
                             'api_base_url': 'https://' + instance}
            for option in self.local_config.config.options(instance):
                instance_data[option] = self.local_config.get(instance, option)
            self.instances.append(instance)
            self.instance_data[instance] = instance_data
        # last used instance
        instance = self.config_store.get('pixelfed', 'instance')
        if not instance:
            return False
        self.client_data = self.instance_data[instance]
        # get user access token
        token = self.get_password()
        if not token:
            return False
        self.user_data['token'] = eval(token)
        return True

    def service_name(self):
        if 'token' in self.user_data:
            # logged in to a particular server
            return self.client_data['name']
        return translate('PixelfedTab', 'Pixelfed')

    def new_session(self, **kw):
        session = PixelfedSession(
            user_data=self.user_data, client_data=self.client_data, **kw)
        session.new_token.connect(self.new_token)
        return session

    def authorise(self):
        dialog = ChooseInstance(
            parent=self.parentWidget(), default=self.client_data['name'],
            instances=self.instances)
        instance = dialog.execute()
        if not instance:
            return
        if not self.register_app(instance):
            return
        self.config_store.set('pixelfed', 'instance', instance)
        super(PixelfedUser, self).authorise()

    def register_app(self, instance):
        if instance in self.instance_data:
            self.client_data = self.instance_data[instance]
            return True
        # create new registration
        api_base_url = 'https://' + instance
        data = {
            'client_name': 'Photini',
            'scopes': ' '.join(self.scopes),
            'redirect_uris': 'http://127.0.0.1',
            'website': 'https://photini.readthedocs.io/',
            }
        rsp = PixelfedSession.check_response(requests.post(
            api_base_url + '/api/v1/apps', data=data, timeout=20))
        if not rsp:
            return False
        client_id = rsp['client_id']
        client_secret = rsp['client_secret']
        # store result
        self.local_config.set(instance, 'client_id', client_id)
        self.local_config.set(instance, 'client_secret', client_secret)
        self.local_config.save()
        self.client_data = {
            'name': instance,
            'api_base_url': api_base_url,
            'client_id': client_id,
            'client_secret': client_secret,
            }
        self.instances.append(instance)
        self.instance_data[instance] = self.client_data
        return True

    def auth_exchange(self, redirect_uri):
        with OAuth2Session(self.client_data['client_id'],
                           redirect_uri=redirect_uri,
                           scope=self.scopes) as session:
            authorization_url, state = session.authorization_url(
                self.client_data['api_base_url'] + '/oauth/authorize')
            result = yield authorization_url
            # often get more scopes than asked for, which upsets
            # requests_oauthlib
            os.environ['OAUTHLIB_RELAX_TOKEN_SCOPE'] = '1'
            token = session.fetch_token(
                self.client_data['api_base_url'] + '/oauth/token',
                client_secret=self.client_data['client_secret'],
                code=result['code'][0])
        if 'access_token' not in token:
            logger.info('No access token received')
            return
        self.new_token(token)
        self.connection_changed.emit(True)

    def unauthorise(self):
        if (self.client_data['version']['pixelfed']
                and self.client_data['version']['pixelfed'] < (0, 12)):
            return
        data = {
            'client_id': self.client_data['client_id'],
            'client_secret': self.client_data['client_secret'],
            'token': self.user_data['token']['access_token']
            }
        PixelfedSession.check_response(requests.post(
            self.client_data['api_base_url'] + '/oauth/revoke', data=data))

    @QtSlot(dict)
    @catch_all
    def new_token(self, token):
        self.user_data['token'] = token
        self.set_password(repr(token))


class ThumbList(QtWidgets.QWidget):
    def __init__(self, *arg, scroll_area=None, **kw):
        super(ThumbList, self).__init__(*arg, **kw)
        self.scroll_area = scroll_area
        self.app = QtWidgets.QApplication.instance()
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().addStretch(1)
        self.image_selection = []
        self.max_media_attachments = 4

    @QtSlot(dict)
    @catch_all
    def new_instance_config(self, config):
        self.max_media_attachments = config[
            'configuration']['statuses']['max_media_attachments']
        self.new_selection(self.app.image_list.get_selected_images())

    def get_selected_images(self):
        return [x[0] for x in self.image_selection]

    def new_selection(self, selection):
        width = self.layout().contentsRect().width()
        n = len(self.image_selection)
        while n:
            n -= 1
            if self.image_selection[n][0] not in selection:
                image, widget = self.image_selection.pop(n)
                self.layout().removeWidget(widget)
                widget.setParent(None)
        selection = selection[:self.max_media_attachments]
        added = False
        for image in selection:
            if image not in self.get_selected_images():
                widget = QtWidgets.QLabel()
                widget.setWordWrap(True)
                widget.setFrameStyle(widget.Shape.Box)
                widget.setStyleSheet('QLabel {background-color: black;}')
                widget.setToolTip(image.name)
                widget.setAlignment(Qt.AlignmentFlag.AlignHCenter |
                                    Qt.AlignmentFlag.AlignVCenter)
                widget.setFixedSize(width, width)
                image.load_thumbnail(label=widget)
                self.layout().insertWidget(self.layout().count() - 1, widget)
                self.image_selection.append((image, widget))
                added = True
        if added and self.scroll_area:
            QtCore.QTimer.singleShot(1, self.scroll_to_last_widget)

    @QtSlot()
    @catch_all
    def scroll_to_last_widget(self):
        scroll_bar = self.scroll_area.verticalScrollBar()
        scroll_bar.setValue(scroll_bar.maximum())


class ScrollArea(QtWidgets.QScrollArea):
    @catch_all
    def resizeEvent(self, event):
        super(ScrollArea, self).resizeEvent(event)
        self.widget().setFixedWidth(self.viewport().contentsRect().width())


class TabWidget(PhotiniUploader):
    logger = logger

    def __init__(self, *arg, **kw):
        self.user_widget = PixelfedUser()
        super(TabWidget, self).__init__(*arg, **kw)

    @staticmethod
    def tab_name():
        return translate('PixelfedTab', '&Pixelfed upload')

    def config_columns(self):
        self.replace_prefs = {'description': True}
        self.upload_prefs = {}
        ## first column
        column = QtWidgets.QGridLayout()
        column.setContentsMargins(0, 0, 0, 0)
        scroll_area = ScrollArea()
        scroll_area.setMinimumWidth(width_for_text(scroll_area, 'x' * 14))
        self.widget['thumb_list'] = ThumbList(scroll_area=scroll_area)
        self.user_widget.new_instance_config.connect(
            self.widget['thumb_list'].new_instance_config)
        scroll_area.setWidget(self.widget['thumb_list'])
        scroll_area.setWidgetResizable(True)
        scroll_area.setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        column.addWidget(scroll_area, 0, 0)
        yield column, 0
        ## second column
        column = QtWidgets.QGridLayout()
        column.setContentsMargins(0, 0, 0, 0)
        group = QtWidgets.QGroupBox()
        group.setLayout(FormLayout(wrapped=True))
        sub_grid = QtWidgets.QGridLayout()
        sub_grid.setContentsMargins(0, 0, 0, 0)
        sub_grid.addWidget(
            Label(translate('PixelfedTab', 'Caption'), layout=group.layout()),
            0, 0)
        self.widget['auto_status'] = QtWidgets.QPushButton(
            translate('PixelfedTab', 'Generate'))
        self.widget['auto_status'].clicked.connect(self.auto_status)
        sub_grid.addWidget(self.widget['auto_status'], 0, 2)
        self.widget['status'] = MultiLineEdit(
            'status', spell_check=True, length_check=1000, length_always=True)
        policy = self.widget['status'].sizePolicy()
        policy.setVerticalStretch(1)
        self.widget['status'].setSizePolicy(policy)
        sub_grid.addWidget(self.widget['status'], 1, 0, 1, 3)
        sub_grid.setColumnStretch(1, 1)
        group.layout().addRow(sub_grid)
        self.widget['spoiler_text'] = SingleLineEdit(
            'spoiler_text', spell_check=True,
            length_check=140, length_always=True)
        self.widget['spoiler_text'].textChanged.connect(self.new_spoiler_text)
        group.layout().addRow(
            translate('PixelfedTab', 'Spoiler'), self.widget['spoiler_text'])
        column.addWidget(group, 0, 0)
        yield column, 1
        ## third column
        column = QtWidgets.QGridLayout()
        column.setContentsMargins(0, 0, 0, 0)
        group = QtWidgets.QGroupBox()
        group.setMinimumWidth(width_for_text(group, 'x' * 23))
        group.setLayout(FormLayout(wrapped=True))
        # visibility
        self.widget['visibility'] = DropDownSelector(
            'visibility', values = (
                (translate('PixelfedTab', 'Public'), 'public'),
                (translate('PixelfedTab', 'Unlisted'), 'unlisted'),
                (translate('PixelfedTab', 'Followers only'), 'private')),
            default='public', with_multiple=False)
        group.layout().addRow(translate('PixelfedTab', 'Post visibility'),
                              self.widget['visibility'])
        # sensitive
        self.widget['sensitive'] = QtWidgets.QCheckBox(
            translate('PixelfedTab', 'Sensitive post'))
        group.layout().addRow(self.widget['sensitive'])
        # comments disabled
        self.widget['comments_disabled'] = QtWidgets.QCheckBox(
            translate('PixelfedTab', 'Disable comments'))
        group.layout().addRow(self.widget['comments_disabled'])
        # licence
        self.widget['license'] = DropDownSelector(
            'license', values=(
                (translate('PixelfedTab', 'All Rights Reserved'), 1),
                (translate('PixelfedTab', 'Public Domain Work'), 5),
                (translate('PixelfedTab',
                           'Public Domain Dedication (CC0)'), 6),
                (translate('PixelfedTab', 'Attribution'), 11),
                (translate('PixelfedTab', 'Attribution-ShareAlike'), 12),
                (translate('PixelfedTab', 'Attribution-NonCommercial'), 13),
                (translate('PixelfedTab',
                           'Attribution-NonCommercial-ShareAlike'), 14),
                (translate('PixelfedTab', 'Attribution-NoDerivs'), 15),
                (translate('PixelfedTab',
                           'Attribution-NonCommercial-NoDerivs'), 16),
                ),
            default=1, with_multiple=False)
        group.layout().addRow(
            translate('PixelfedTab', 'Licence'), self.widget['license'])
        column.addWidget(group, 0, 0)
        # create new collection
        self.widget['new_album'] = QtWidgets.QPushButton(
            translate('PixelfedTab', 'New collection'))
        self.widget['new_album'].clicked.connect(self.new_album)
        column.addWidget(self.widget['new_album'], 1, 0)
        # update alt text
        self.widget['sync_remote'] = QtWidgets.QPushButton(
            translate('PixelfedTab', 'Update remote'))
        self.widget['sync_remote'].clicked.connect(self.sync_remote)
        column.addWidget(self.widget['sync_remote'], 2, 0)
        yield column, 0
        ## last column is list of albums
        yield self.album_list(
            label=translate('PixelfedTab', 'Add to collections'),
            max_selected=3), 0

    @QtSlot()
    @catch_all
    def new_spoiler_text(self):
        if self.widget['spoiler_text'].get_value():
            self.widget['sensitive'].setChecked(True)
            self.widget['sensitive'].setEnabled(False)
        else:
            self.widget['sensitive'].setEnabled(True)

    def accepted_image_type(self, file_type):
        return file_type in self.user_widget.instance_config[
            'configuration']['media_attachments']['supported_mime_types']

    def process_image(self, image):
        image = self.read_image(image)
        image = self.data_to_image(image)
        # reduce image size
        w, h = image['width'], image['height']
        shrink = math.sqrt(float(w * h) /
                           float(self.user_widget.max_size['image']['pixels']))
        if shrink > 1.0:
            w = int(float(w) / shrink)
            h = int(float(h) / shrink)
            logger.info('Resizing %s to %dx%d pixels', image['name'], w, h)
            image = self.resize_image(image, w, h)
        # convert to jpeg and reduce size if necessary
        image = self.image_to_data(
            image, mime_type='image/jpeg',
            max_size=self.user_widget.max_size['image']['bytes'])
        return image['data'], image['mime_type']

    def get_selected_images(self):
        return self.widget['thumb_list'].get_selected_images()

    @QtSlot()
    @catch_all
    def sync_remote(self):
        image_list = {}
        for image in self.app.image_list.get_selected_images():
            if not image.metadata.keywords:
                continue
            for keyword, (ns, predicate,
                          value) in image.metadata.keywords.machine_tags():
                if (ns == self.user_widget.client_data['name']
                        and predicate != 'id'):
                    image_list[value] = image
        if not image_list:
            return
        with Busy():
            with self.user_widget.session(parent=self) as session:
                for media_id, image in image_list.items():
                    data = {'description': self.alt_text(image)}
                    session.api_call(
                        '/api/v1/media/{}'.format(media_id), 'PUT', data=data)

    def alt_text(self, image):
        description = []
        lang = self.user_widget.user_data['lang']
        max_len = int(self.user_widget.compose_settings['max_altext_length'])
        if image.metadata.alt_text:
            text = image.metadata.alt_text.best_match(lang)
            description.append(text)
            max_len -= len(text) + 2
        if image.metadata.alt_text_ext:
            text = image.metadata.alt_text_ext.best_match(lang)
            if len(text) <= max_len:
                description.append(text)
        if description:
            return '\n\n'.join(description)
        return ''

    def get_upload_params(self, image, state):
        if 'ask_alt_text' not in state:
            state['ask_alt_text'] = self.user_widget.compose_settings[
                                                        'media_descriptions']
        params = {'media': {}, 'status': {}}
        params['file_name'] = os.path.basename(image.path)
        # 'description' is the ALT text for an image
        params['media']['description'] = self.alt_text(image)
        if state['ask_alt_text'] and not params['media']['description']:
            dialog = QtWidgets.QMessageBox(parent=self)
            dialog.setWindowTitle(
                translate('PixelfedTab', 'Photini: alt text'))
            dialog.setText('<h3>{}</h3>'.format(
                translate('PixelfedTab', 'Missing alt text.')))
            dialog.setInformativeText(translate(
                'PixelfedTab', 'File "{file_name}" does not have any'
                ' "alt text" for accessibility.'
                ' Would you like to upload it anyway?').format(
                    file_name=os.path.basename(image.path)))
            dialog.setStandardButtons(dialog.StandardButton.Yes |
                                      dialog.StandardButton.YesToAll |
                                      dialog.StandardButton.Abort)
            dialog.setIcon(dialog.Icon.Warning)
            result = execute(dialog)
            if result == dialog.StandardButton.YesToAll:
                state['ask_alt_text'] = False
            elif result == dialog.StandardButton.Abort:
                return 'abort'
        focus = image.metadata.image_region.get_focus(image)
        if focus:
            params['media']['focus'] = '{:f},{:f}'.format(*focus)
        params['license'] = self.widget['license'].get_value()
        params['default_license'] = int(
            self.user_widget.compose_settings['default_license'])
        # 'status' is the text that accompanies the media
        for key in ('status', 'spoiler_text', 'visibility'):
            params['status'][key] = self.widget[key].get_value()
        params['status']['status'] = params[
            'status']['status'] or params['file_name']
        params['status']['sensitive'] = ('0', '1')[
            self.widget['sensitive'].isChecked()]
        if not self.user_widget.unavailable['comments_disabled']:
            params['status']['comments_disabled'] = ('0', '1')[
                self.widget['comments_disabled'].isChecked()]
        if not self.user_widget.unavailable['albums']:
            # collections to add post to
            album_ids = self.widget['albums'].get_checked_ids()
            if album_ids:
                params['status']['collection_ids[]'] = album_ids
        return params

    def enable_upload_button(self, selection=None):
        selection = selection or self.app.image_list.get_selected_images()
        super(TabWidget, self).enable_upload_button(selection=selection)
        if (self.buttons['upload'].isEnabled()
                and len(selection) > self.widget[
                    'thumb_list'].max_media_attachments
                and not self.buttons['upload'].is_checked()):
            self.buttons['upload'].setEnabled(False)
        self.widget['thumb_list'].new_selection(selection)
        enabled = self.widget['status'].isEnabled()
        self.widget['sync_remote'].setEnabled(enabled and bool(selection))
        self.widget['auto_status'].setEnabled(
            enabled and (len(selection) > 0 and len(selection) <=
                         self.widget['thumb_list'].max_media_attachments))

    @QtSlot()
    @catch_all
    def auto_status(self):
        result = {
            'title': [], 'headline': [], 'description': [], 'keywords': []}
        for image in self.get_selected_images():
            for key in result:
                result[key].append(getattr(image.metadata, key))
        lang = self.user_widget.user_data['lang']
        result['title'] = [x.best_match(lang) for x in result['title'] if x]
        result['headline'] = [str(x) for x in result['headline'] if x]
        result['description'] = [
            x.best_match(lang) for x in result['description'] if x]
        result['keywords'] = [x.human_tags() for x in result['keywords'] if x]
        result['keywords'] = list(itertools.chain(*result['keywords']))
        strings = []
        for key in ('title', 'headline', 'description'):
            if not result[key]:
                continue
            string = result[key][0]
            for text in result[key][1:]:
                if string in text:
                    string = text
                elif text not in string:
                    string += '\n' + text
            strings.append(string)
        if result['keywords']:
            # remove punctuation
            keywords = [re.sub(r'[^\w\s]', '', x, re.UNICODE)
                        for x in result['keywords']]
            # convert to #CamelCase
            keywords = ['#' + x.title().replace(' ', '') for x in keywords]
            strings.append(' '.join(set(keywords)))
        self.widget['status'].set_value('\n\n'.join(strings))

    @QtSlot()
    @catch_all
    def new_album(self):
        dialog = QtWidgets.QDialog(parent=self)
        dialog.setWindowTitle(translate('PixelfedTab', 'Create new collection'))
        dialog.setLayout(FormLayout())
        title = SingleLineEdit('title', spell_check=True,
                               length_check=50, length_always=True)
        dialog.layout().addRow(translate('PixelfedTab', 'Title'), title)
        description = MultiLineEdit('description', spell_check=True,
                                    length_check=500, length_always=True)
        dialog.layout().addRow(
            translate('PixelfedTab', 'Description'), description)
        visibility = DropDownSelector(
            'visibility', values = (
                (translate('PixelfedTab', 'Public'), 'public'),
                (translate('PixelfedTab', 'Followers only'), 'private'),
                (translate('PixelfedTab', 'Draft'), 'draft')),
            default='public', with_multiple=False)
        dialog.layout().addRow(
            translate('PixelfedTab', 'Collection visibility'), visibility)
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok |
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(dialog.accept)
        button_box.rejected.connect(dialog.reject)
        dialog.layout().addRow(button_box)
        if execute(dialog) != QtWidgets.QDialog.DialogCode.Accepted:
            return
        data = {
            'title': title.toPlainText(),
            'description': description.toPlainText(),
            'visibility': visibility.get_value(),
            }
        if not data['title']:
            return
        # currently no known endpoint to create a collection
        return
        with self.user_widget.session(parent=self) as session:
            # this endpoint doesn't exist
            album = session.api_call('/api/v1.1/collections/create')
            # this endpoint updates an existing album
            album = session.api_call(
                '/api/v1.1/collections/update/{}'.format(
                    album['id']), data=data)
        widget = self.widget['albums'].add_album(
            {'title': album['title'], 'description': album['description'],
             'id': album['id'], 'writeable': True},
            index=0)
        widget.setChecked(True)



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2018-22  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import locale
import logging

import requests

from photini.configstore import key_store
from photini.photinimap import GeocoderBase, PhotiniMap
from photini.pyqt import Busy, catch_all, QtCore, QtGui, QtSlot, QtWidgets
from photini.widgets import CompactButton

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class MapboxGeocoder(GeocoderBase):
    api_key = key_store.get('mapboxmap', 'api_key')
    cache_size = 0

    def query(self, params):
        query = params['query']
        del params['query']
        params['access_token'] = self.api_key
        params['autocomplete '] = 'false'
        lang, encoding = locale.getdefaultlocale()
        if lang:
            params['language'] = lang
        query += '.json'
        url = 'https://api.mapbox.com/geocoding/v5/mapbox.places/' + query
        with Busy():
            self.rate_limit()
            try:
                rsp = requests.get(url, params=params, timeout=5)
            except Exception as ex:
                logger.error(str(ex))
                return []
        if rsp.status_code >= 400:
            logger.error('Search error %d', rsp.status_code)
            return []
        self.block_timer.setInterval(
            self.interval * 600 // max(int(rsp.headers['X-Rate-Limit-Limit']), 1))
        rsp = rsp.json()
        return rsp['features']

    def search(self, search_string, bounds=None):
        params = {
            'query': search_string,
            'limit': '10',
            }
        if bounds:
            north, east, south, west = bounds
            margin = (10.0 + west - east) / 2.0
            if margin > 0.0:
                east += margin
                west -= margin
            margin = (10.0 + south - north) / 2.0
            if margin > 0.0:
                north = min(north + margin,  90.0)
                south = max(south - margin, -90.0)
            params['bbox'] = '{:.4f},{:.4f},{:.4f},{:.4f}'.format(
                west, south, east, north)
        for feature in self.query(params):
            if 'place_name' not in feature:
                continue
            if 'bbox' in feature:
                west, south, east, north = feature['bbox']
                yield north, east, south, west, feature['place_name']
            elif 'center' in feature:
                east, north = feature['center']
                yield north, east, None, None, feature['place_name']

    def search_terms(self):
        widget = CompactButton(
            translate('MapTabMapbox', 'Search powered by Mapbox'))
        widget.clicked.connect(self.load_mapbox_tos)
        return [widget]

    @QtSlot()
    @catch_all
    def load_mapbox_tos(self):
        QtGui.QDesktopServices.openUrl(
            QtCore.QUrl('https://www.mapbox.com/tos/'))


class TabWidget(PhotiniMap):
    api_key = key_store.get('mapboxmap', 'api_key')

    @staticmethod
    def tab_name():
        return translate('MapTabMapbox', 'Map (&Mapbox)')

    def get_geocoder(self):
        return MapboxGeocoder(parent=self)

    def get_head(self):
        return '''    <link rel="stylesheet" href="{url}/mapbox.css" />
    <script type="text/javascript" src="{url}/mapbox.js">
    </script>
    <script type="text/javascript">
      L.mapbox.accessToken = "{key}";
    </script>'''.format(key=self.api_key,
                        url='https://api.mapbox.com/mapbox.js/v3.3.1')



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2019-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import logging
import os
import urllib

import requests
from requests_oauthlib import OAuth2Session

from photini.pyqt import (
    catch_all, execute, QtCore, QtSlot, QtWidgets, width_for_text)
from photini.uploader import PhotiniUploader, UploaderSession, UploaderUser

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate

# Google Photos API: https://developers.google.com/photos/library/reference/rest

class GooglePhotosSession(UploaderSession):
    oauth_url  = 'https://www.googleapis.com/oauth2/'
    photos_url = 'https://photoslibrary.googleapis.com/'

    def open_connection(self):
        if self.api:
            return
        self.api = OAuth2Session(
            client_id=self.client_data['client_id'],
            token=self.user_data['token'],
            auto_refresh_url=self.oauth_url + 'v4/token',
            auto_refresh_kwargs=self.client_data, token_updater=self.save_token)
        self.api.headers.update(self.headers)

    def save_token(self, token):
        self.user_data['token'] = token
        self.new_token.emit(token)

    def api_call(self, url, post=False, **params):
        self.open_connection()
        if post:
            rsp = self.api.post(url, timeout=5, **params)
        else:
            rsp = self.api.get(url, timeout=5, **params)
        rsp = self.check_response(rsp)
        if not rsp:
            self.close_connection()
        return rsp

    def new_album(self, title):
        body = {'album': {'title': title}}
        return self.api_call(
            self.photos_url + 'v1/albums', json=body, post=True)

    def do_upload(self, fileobj, image_type, image, params):
        # see https://developers.google.com/photos/library/guides/upload-media
        # 1/ initiate a resumable upload session (to do file in chunks)
        headers = {
            'X-Goog-Upload-Command'     : 'start',
            'X-Goog-Upload-Content-Type': image_type,
            'X-Goog-Upload-File-Name'   : os.path.basename(image.path),
            'X-Goog-Upload-Protocol'    : 'resumable',
            'X-Goog-Upload-Raw-Size'    : str(fileobj.len),
            }
        rsp = self.api.post(self.photos_url + 'v1/uploads', headers=headers)
        rsp.raise_for_status()
        upload_url = rsp.headers['X-Goog-Upload-URL']
        chunk_size = int(rsp.headers['X-Goog-Upload-Chunk-Granularity'])
        # 2/ upload data in chunks, size set by google
        self.upload_progress.emit({'busy': False})
        headers = {'X-Goog-Upload-Command': 'upload'}
        offset = 0
        while offset < fileobj.len:
            chunk = fileobj.read(chunk_size)
            headers['X-Goog-Upload-Offset'] = str(offset)
            offset += len(chunk)
            if offset >= fileobj.len:
                headers['X-Goog-Upload-Command'] = 'upload, finalize'
            rsp = self.api.post(upload_url, headers=headers, data=chunk)
            rsp.raise_for_status()
            self.upload_progress.emit({'value': offset * 100 // fileobj.len})
            if rsp.text:
                upload_token = rsp.text
        self.upload_progress.emit({'busy': True})
        if not upload_token:
            return 'no upload token received'
        # 3/ convert uploaded bytes to a media item
        body = {'newMediaItems': [{
            'description'    : params['description'],
            'simpleMediaItem': {'uploadToken': upload_token},
            }]}
        if params['albums']:
            body['albumId'] = params['albums'][0]
        rsp = self.api_call(self.photos_url + 'v1/mediaItems:batchCreate',
                            json=body, post=True)
        if not (rsp and 'newMediaItemResults' in rsp):
            return 'failed to create media item'
        rsp = rsp['newMediaItemResults'][0]
        if 'status' in rsp:
            if rsp['status']['message'] not in ('Success', 'OK'):
                return str(rsp['status'])
        media_id = rsp['mediaItem']['id']
        # 4/ add media item to more albums
        if len(params['albums']) > 1:
            body = {'mediaItemIds': [media_id]}
            for album_id in params['albums'][1:]:
                url = (self.photos_url + 'v1/albums/' +
                       album_id + ':batchAddMediaItems')
                rsp = self.api_call(url, json=body, post=True)
        return ''


class GooglePhotosUser(UploaderUser):
    logger = logger
    config_section = 'googlephotos'
    scope = ('profile', 'https://www.googleapis.com/auth/photoslibrary')
    max_size = {'image': {'bytes': 200 * (2 ** 20)},
                'video': {'bytes': 10 * (2 ** 30)}}

    def on_connect(self, widgets):
        with self.session(parent=self) as session:
            # check auth
            yield 'connected', session.api.authorized
            # get user details
            rsp = session.api_call(session.oauth_url + 'v2/userinfo')
            if not rsp:
                yield 'connected', False
            self.user_data['lang'] = rsp['locale']
            name = rsp['name']
            rsp = session.check_response(
                session.api.get(rsp['picture']), decode=False)
            picture = rsp and rsp.content
            yield 'user', (name, picture)
            # get albums
            params = {}
            while True:
                rsp = session.api_call(
                    session.photos_url + 'v1/albums', params=params)
                if not (rsp and 'albums' in rsp):
                    break
                for album in rsp['albums']:
                    if 'id' not in album:
                        logger.info('Malformed album', album)
                        continue
                    yield 'album', self.normalise_album(album)
                if 'nextPageToken' not in rsp:
                    break
                params['pageToken'] = rsp['nextPageToken']

    def load_user_data(self):
        refresh_token = self.get_password()
        if not refresh_token:
            return False
        # create an expired token
        self.user_data['token'] = {
            'access_token' : 'xxx',
            'refresh_token': refresh_token,
            'expires_in'   : -30,
            }
        return True

    @staticmethod
    def service_name():
        return translate('GooglePhotosTab', 'Google Photos')

    def new_session(self, **kw):
        session = GooglePhotosSession(
            user_data=self.user_data, client_data=self.client_data, **kw)
        session.new_token.connect(self.new_token)
        return session

    def auth_exchange(self, redirect_uri):
        code_verifier = ''
        while len(code_verifier) < 43:
            code_verifier += OAuth2Session().new_state()
        auth_params = {
            'code_verifier': code_verifier,
            'redirect_uri' : redirect_uri,
            }
        auth_params.update(self.client_data)
        url = 'https://accounts.google.com/o/oauth2/v2/auth'
        url += '?client_id=' + auth_params['client_id']
        url += '&redirect_uri=' + auth_params['redirect_uri']
        url += '&response_type=code'
        url += '&scope=' + urllib.parse.quote(' '.join(self.scope))
        url += '&code_challenge=' + auth_params['code_verifier']
        result = yield url
        if not 'code' in result:
            logger.info('No authorisaton code received')
            return
        data = {
            'code'      : result['code'][0],
            'grant_type': 'authorization_code',
            }
        data.update(auth_params)
        try:
            rsp = requests.post(GooglePhotosSession.oauth_url + 'v4/token',
                                data=data, timeout=5)
            rsp.raise_for_status()
            rsp = rsp.json()
        except Exception as ex:
            logger.error(str(ex))
            return
        if 'access_token' not in rsp:
            logger.info('No access token received')
            return
        self.new_token(rsp)
        self.connection_changed.emit(True)

    @QtSlot(dict)
    @catch_all
    def new_token(self, token):
        self.user_data['token'] = token
        self.set_password(self.user_data['token']['refresh_token'])

    @staticmethod
    def normalise_album(album):
        album['writeable'] = 'isWriteable' in album and album['isWriteable']
        album['description'] = None
        return album


class TabWidget(PhotiniUploader):
    logger = logger

    def __init__(self, *arg, **kw):
        self.user_widget = GooglePhotosUser()
        super(TabWidget, self).__init__(*arg, **kw)

    @staticmethod
    def tab_name():
        return translate('GooglePhotosTab', 'Google &Photos upload')

    def config_columns(self):
        ## first column
        column = QtWidgets.QGridLayout()
        column.setContentsMargins(0, 0, 0, 0)
        # create new set
        group = QtWidgets.QGroupBox()
        group.setMinimumWidth(width_for_text(group, 'x' * 23))
        column.addWidget(group, 0, 0)
        new_set_button = QtWidgets.QPushButton(
            translate('GooglePhotosTab', 'New album'))
        new_set_button.clicked.connect(self.new_set)
        column.addWidget(new_set_button, 1, 0)
        column.setRowStretch(0, 1)
        yield column, 0
        ## last column is list of albums
        yield self.album_list(), 1

    def accepted_image_type(self, file_type):
        # see https://developers.google.com/photos/library/guides/upload-media#file-types-sizes
        return file_type in ('image/gif', 'image/heic', 'image/jpeg',
                             'image/png', 'image/tiff', 'image/webp',
                             'image/x-ms-bmp')

    def get_conversion_function(self, image, state, params):
        convert = super(
            TabWidget, self).get_conversion_function(image, state, params)
        if convert == 'omit':
            return convert
        # Google's docs say to remind user of storage limits if uploads
        # exceed 25 MB per user. We do it if any item exceeds 25 MB.
        max_size = 25 * 1024 * 1024
        size = os.stat(image.path).st_size
        if size <= max_size:
            return convert
        dialog = QtWidgets.QMessageBox(parent=self)
        dialog.setWindowTitle(
            translate('GooglePhotosTab', 'Photini: large file'))
        dialog.setText('<h3>{}</h3>'.format(
            translate('GooglePhotosTab', 'Large file.')))
        dialog.setInformativeText(
            translate('GooglePhotosTab',
                      'File "{file_name}" is over 25 MB. Remember that Photini '
                      'uploads count towards storage in your Google Account. '
                      'Upload it anyway?').format(
                          file_name=os.path.basename(image.path)))
        dialog.setIcon(dialog.Icon.Warning)
        dialog.setStandardButtons(dialog.StandardButton.Yes)
        self.add_skip_button(dialog)
        if execute(dialog) == dialog.StandardButton.Yes:
            return convert
        return 'omit'

    def get_upload_params(self, image, state):
        lang = self.user_widget.user_data['lang']
        description = []
        if image.metadata.title:
            description.append(image.metadata.title.best_match(lang))
        if image.metadata.headline:
            description.append(image.metadata.headline)
        if image.metadata.description:
            description.append(image.metadata.description.best_match(lang))
        if description:
            description = '\n\n'.join(description)
        else:
            description = image.path
        params = {
            'description': description,
            'albums'     : self.widget['albums'].get_checked_ids(),
            }
        return params

    @QtSlot()
    @catch_all
    def new_set(self):
        title, OK = QtWidgets.QInputDialog.getText(
            self, translate('GooglePhotosTab', 'Album title'),
            translate('GooglePhotosTab', 'Please enter a title for the album'))
        if not OK or not title:
            return
        with self.user_widget.session(parent=self) as session:
            album = session.new_album(title)
        if not album:
            return
        album = GooglePhotosUser.normalise_album(album)
        widget = self.widget['albums'].add_album(album, index=0)
        widget.setChecked(True)
